# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import base64
import os
from logging import getLogger

import flask
import sqlalchemy.exc

from marv_webapi import webapi
from ..model import db


log = getLogger(__name__)


class DBNotInitialized(Exception):
    pass


def create_app(site, config_obj=None, app_root=None, init=None, **kw):
    app = flask.Flask(__name__)
    app.site = site

    # default config
    app_root = app_root.rstrip('/') if app_root else None
    app.config['APPLICATION_ROOT'] = app_root or None
    app.config['JSONIFY_PRETTYPRINT_REGULAR'] = False
    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0
    app.config['SQLALCHEMY_DATABASE_URI'] = site.config.marv.dburi
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    if config_obj is not None:
        app.config.from_object(config_obj)
    app.config.update(kw)

    db.init_app(app)
    with app.app_context():
        if init:
            site.init()
        try:
            db.session.execute('SELECT name FROM sqlite_master WHERE type="table";')
        except sqlalchemy.exc.OperationalError:
            if init is None:  # auto-init
                site.init()
                db.session.execute('SELECT name FROM sqlite_master WHERE type="table";')
            else:
                raise DBNotInitialized()

        app.acl = site.config.marv.acl()
        webapi.init_app(app, url_prefix='/marv/api')

    with open(site.config.marv.sessionkey_file) as f:
        app.config['SECRET_KEY'] = f.read()

    staticdir = site.config.marv.staticdir
    with open(os.path.join(staticdir, 'index.html')) as f:
        index_html = f.read().replace('MARV_APP_ROOT', app_root or "")

    customjs = os.path.join(site.config.marv.frontenddir, 'custom.js')
    try:
        with(open(customjs)) as f:
            data = base64.b64encode(f.read())
    except IOError:
        pass
    else:
        assert '<script async src="main-built.js"></script>' in index_html
        index_html = index_html.replace(
            '<script async src="main-built.js"></script>',
            '<script src="data:text/javascript;base64,{}"></script>'.format(data) +
            '\n<script async src="main-built.js"></script>', 1
        )

    customcss = os.path.join(site.config.marv.frontenddir, 'custom.css')
    try:
        with(open(customcss)) as f:
            data = base64.b64encode(f.read())
    except IOError:
        pass
    else:
        assert '<link async rel="stylesheet" href="main-built.css" />' in index_html
        index_html = index_html.replace(
            '<link async rel="stylesheet" href="main-built.css" />',
            '<link async rel="stylesheet" href="main-built.css" />' +
            '<link rel="stylesheet" href="data:text/css;base64,{}" />'.format(data), 1
        )

    customdir = os.path.join(site.config.marv.frontenddir, 'custom')
    @app.route('/custom/<path:path>')
    def custom(path):
        resp = flask.send_from_directory(customdir, path, conditional=True)
        resp.headers['Cache-Control'] = 'no-cache'
        return resp

    @app.route('/', defaults={'path': ''})
    @app.route('/<path:path>')
    def assets(path):
        if not path:
            path = 'index.html'

        if path == 'index.html':
            resp = flask.make_response(index_html)
            resp.headers['Cache-Control'] = 'no-cache'
            return resp

        if path == 'docs':
            return flask.redirect(flask.request.base_url + '/', 301)

        if path == 'docs/':
            path = 'docs/index.html'

        resp = flask.send_from_directory(staticdir, path, conditional=True)
        resp.headers['Cache-Control'] = 'no-cache'
        return resp

    return app



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import unittest

from ..run import run_nodes
from ..testing import make_dataset, make_sink, marv

dataset = make_dataset()
SETID = dataset.setid


class TestCase(unittest.TestCase):
    def test_foreach_cascade(self):
        @marv.node()
        @marv.input('stream', foreach=[0, 10])
        def images(stream):
            """Produce 2 streams each with two messages"""
            yield marv.set_header(title=stream)
            yield marv.push(1+stream)
            yield marv.push(2+stream)
            yield marv.push(3+stream)
            yield marv.push(4+stream)
            yield marv.push(5+stream)

        @marv.node()
        @marv.input('stream', foreach=images)
        def galleries(stream):
            """Consume each stream into a list"""
            yield marv.set_header(title=stream.title)  # TODO: This is currently needed
            images = []
            while True:
                img = yield marv.pull(stream)
                if img is None:
                    break
                images.append(img)
            yield marv.push({'images': images})

        @marv.node()
        @marv.input('galleries', default=galleries)
        def images_section(galleries):
            """Consume both galleries into a list"""
            tmp = []
            while True:
                msg = yield marv.pull(galleries)
                if msg is None:
                    break
                tmp.append(msg)
            galleries = tmp
            galleries = yield marv.pull_all(*galleries)
            yield marv.push({'galleries': galleries})

        sink = make_sink(images_section)
        run_nodes(dataset, [sink], {})
        self.assertEqual(sink.stream, [{'galleries': [{'images': [1, 2, 3, 4, 5]},
                                                      {'images': [11, 12, 13, 14, 15]}]}])



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import sys
from pkg_resources import iter_entry_points

from marv_node.io import create_group
from marv_node.io import create_stream
from marv_node.io import fork
from marv_node.io import get_logger
from marv_node.io import get_requested
from marv_node.io import get_stream
from marv_node.io import make_file
from marv_node.io import pull
from marv_node.io import pull_all
from marv_node.io import pull_any
from marv_node.io import push
from marv_node.io import set_header
from marv_node.io import Pull
from marv_node.node import input, node
from marv_node.node import Node
from marv_node.tools import select
from marv_webapi.tooling import api_endpoint
from marv_webapi.tooling import api_group

__all__ = [
    'create_group',
    'create_stream',
    'get_logger',
    'get_stream',
    'input',
    'make_file',
    'node',
    'pull',
    'pull_all',
    'push',
    'select',
    'set_header',
]

MODULE = sys.modules[__name__]
for ep in iter_entry_points(group='marv_deco'):
    assert not hasattr(MODULE, ep.name)
    setattr(MODULE, ep.name, ep.load())
del MODULE

from marv_node.io import Abort



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import datetime
import json
import os
import re
import sys
import warnings
from logging import getLogger

import click
from jinja2 import Template
from sqlalchemy.orm.exc import NoResultFound

import marv.app
from marv.config import ConfigError
from marv.model import Comment, Dataset, User, Group, dataset_tag, db
from marv.model import STATUS_MISSING
from marv.site import Site, UnknownNode, dump_database, make_config
from marv.utils import find_obj
from marv_cli import marv as marvcli
from marv_cli import PDB
from marv_node.setid import SetID
from marv_node.stream import RequestedMessageTooOld
from marv_store import DirectoryAlreadyExists


log = getLogger(__name__)
warnings.simplefilter('always', DeprecationWarning)


class NodeParamType(click.ParamType):
    name = 'NODE'

    def convert(self, value, param, ctx):
        return find_obj(value)

NODE = NodeParamType()


def create_app(push=True, init=None):
    ctx = click.get_current_context()
    siteconf = ctx.obj
    if siteconf is None:
        ctx.fail('Could not find config file: ./marv.conf or /etc/marv/marv.conf.\n'
                 'Change working directory or specify explicitly:\n\n'
                 '  marv --config /path/to/marv.conf\n')
    site = Site(siteconf)
    try:
        app = marv.app.create_app(site, init=init)
    except ConfigError as e:
        click.echo('Error {}'.format(e.args[0]), err=True)
        click.get_current_context().exit(1)
    except OSError as e:
        if e.errno == 13:
            print(e, file=sys.stderr)
            sys.exit(13)
        raise
    if push:
        appctx = app.app_context()
        appctx.push()
        ctx.call_on_close(appctx.pop)
    return app


def parse_setids(datasets, discarded=False, dbids=False):
    fail = click.get_current_context().fail
    setids = set()
    for prefix in datasets:
        many = prefix.endswith('*')
        prefix = prefix[:-1] if many else prefix
        setid = db.session.query(Dataset.id if dbids else Dataset.setid)\
                          .filter(Dataset.setid.like('{}%'.format(prefix)))\
                          .filter(Dataset.discarded.is_(discarded))\
                          .all()
        if len(setid) == 0:
            fail('{} does not match any {}dataset'
                 .format(prefix, 'discarded ' if discarded else ''))
        elif len(setid) > 1 and not many:
            matches = '\n  '.join([x[0] for x in setid])
            fail("{} matches multiple:\n  {}\nUse '{}*' to mean all these."
                 .format(prefix, matches, prefix))
        setids.update(x[0] for x in setid)
        # TODO: complain about multiple
    return sorted(setids)


@marvcli.command('cleanup')
@click.option('--discarded/--no-discarded', help='Delete discarded datasets')
@click.option('--unused-tags/--no-unused-tags',
              help='Cleanup unused tags and other relations')
@click.pass_context
def marvcli_cleanup(ctx, discarded, unused_tags):
    """Cleanup unused tags and discarded datasets."""
    if not any([discarded, unused_tags]):
        click.echo(ctx.get_help())
        ctx.exit(1)

    site = create_app().site

    if discarded:
        site.cleanup_discarded()

    if unused_tags:
        site.cleanup_tags()
        site.cleanup_relations()

    # TODO: cleanup unused store paths / unused generations


@marvcli.group('develop')
def marvcli_develop():
    """Development tools."""


@marvcli_develop.command('server')
@click.option('--port', default=5000, help='Port to listen on')
@click.option('--public/--no-public',
              help='Listen on all IPs instead of only 127.0.0.1')
def marvcli_develop_server(port, public):
    """Run development webserver.

    ATTENTION: By default it is only served on localhost. To run it
    within a container and access it from the outside, you need to
    forward the port and tell it to listen on all IPs instead of only
    localhost.
    """
    from flask_cors import CORS
    app = create_app(push=False)
    app.site.load_for_web()
    CORS(app)

    class PDBMiddleware(object):
        def __init__(self, app):
            self.app = app

        def __call__(self, environ, start_response):
            from marv_cli import launch_pdb_on_exception
            with launch_pdb_on_exception():
                appiter = self.app(environ, start_response)
                for item in appiter:
                    yield item

    app.debug = True
    if PDB:
        app.wsgi_app = PDBMiddleware(app.wsgi_app)
        app.run(use_debugger=False,
                use_reloader=False,
                host=('0.0.0.0' if public else '127.0.0.1'),
                port=port,
                threaded=False)
    else:
        app.run(host=('0.0.0.0' if public else '127.0.0.1'),
                port=port,
                reloader_type='watchdog',
                threaded=False)


@marvcli.command('discard')
@click.option('--node', 'nodes', multiple=True, help='TODO: Discard output of selected nodes')
@click.option('--all-nodes', help='TODO: Discard output of all nodes')
@click.option('--comments/--no-comments', help='Delete comments')
@click.option('--tags/--no-tags', help='Delete tags')
@click.option('--confirm/--no-confirm', default=True, show_default=True,
              help='Ask for confirmation before deleting tags and comments')
@click.argument('datasets', nargs=-1, required=True)
def marvcli_discard(datasets, all_nodes, nodes, tags, comments, confirm):
    """Mark DATASETS to be discarded or discard associated data.

    Without any options the specified datasets are marked to be
    discarded via `marv cleanup --discarded`. Use `marv undiscard` to
    undo this operation.

    Otherwise, selected data associated with the specified datasets is
    discarded right away.
    """
    mark_discarded = not any([all_nodes, nodes, tags, comments])

    site = create_app().site
    setids = parse_setids(datasets)

    if tags or comments:
        if confirm:
            msg = ' and '.join(filter(None, ['tags' if tags else None,
                                             'comments' if comments else None]))
            click.echo('About to delete {}'.format(msg))
            click.confirm('This cannot be undone. Do you want to continue?', abort=True)

        ids = [x[0] for x in db.session.query(Dataset.id).filter(Dataset.setid.in_(setids))]
        if tags:
            where = dataset_tag.c.dataset_id.in_(ids)
            stmt = dataset_tag.delete().where(where)
            db.session.execute(stmt)

        if comments:
            comment_table = Comment.__table__
            where = comment_table.c.dataset_id.in_(ids)
            stmt = comment_table.delete().where(where)
            db.session.execute(stmt)

    if nodes or all_nodes:
        storedir = site.config.marv.storedir
        for setid in setids:
            setdir = os.path.join(storedir, setid)
        # TODO: see where we are getting with dep tree tables

    if mark_discarded:
        dataset = Dataset.__table__
        stmt = dataset.update()\
                      .where(dataset.c.setid.in_(setids))\
                      .values(discarded=True)
        db.session.execute(stmt)

    db.session.commit()


@marvcli.command('undiscard')
@click.argument('datasets', nargs=-1, required=True)
def marvcli_undiscard(datasets):
    """Undiscard DATASETS previously discarded."""
    create_app()

    setids = parse_setids(datasets, discarded=True)
    dataset = Dataset.__table__
    stmt = dataset.update()\
                  .where(dataset.c.setid.in_(setids))\
                  .values(discarded=False)
    db.session.execute(stmt)
    db.session.commit()


@marvcli.command('dump')
@click.argument('dump_file', type=click.File('w'))
def marvcli_dump(dump_file):
    """Dump database to json file.

    Use '-' for stdout.
    """
    ctx = click.get_current_context()
    if ctx.obj is None:
        ctx.fail('Could not find config file: ./marv.conf or /etc/marv/marv.conf.\n'
                 'Change working directory or specify explicitly:\n\n'
                 '  marv --config /path/to/marv.conf\n')
    siteconf = make_config(ctx.obj)
    dump = dump_database(siteconf.marv.dburi)
    json.dump(dump, dump_file, sort_keys=True, indent=2)


@marvcli.command('restore')
@click.argument('dump_file', type=click.File())
def marvcli_restore(dump_file):
    """Restore previously dumped database from file.

    Use '-' to read from stdin.
    """
    data = json.load(dump_file)
    site = create_app().site
    site.restore_database(**data)


@marvcli.command('init')
def marvcli_init():
    """(Re-)initialize marv site according to config"""
    create_app(init=True)


@marvcli.command('query')
@click.option('--list-tags', is_flag=True, help='List all tags')
@click.option('--col', '--collection', 'collections', multiple=True,
              help='Limit to one or more collections or force listing of all with --collection=*')
@click.option('--discarded/--no-discarded', help='Dataset is discarded')
@click.option('--missing', is_flag=True, help='Datasets with missing files')
@click.option('--outdated', is_flag=True, help='Datasets with outdated node output')
@click.option('--path', type=click.Path(resolve_path=True),
              help='Dataset contains files whose path starts with PATH')
@click.option('--tagged', 'tags', multiple=True, help='Match any given tag')
@click.option('-0', '--null', is_flag=True, help='Use null byte to separate output instead of newlines')
@click.pass_context
def marvcli_query(ctx, list_tags, collections, discarded, missing, outdated, path, tags, null):
    """Query datasets.

    Use --col=* to list all datasets across all collections.
    """
    if not any([collections, discarded, list_tags, missing, outdated, path, tags]):
        click.echo(ctx.get_help())
        ctx.exit(1)

    sep = '\x00' if null else '\n'

    site = create_app().site

    if '*' in collections:
        collections = None
    else:
        for col in collections:
            if col not in site.collections:
                ctx.fail('Unknown collection: {}'.format(col))

    if list_tags:
        tags = site.listtags(collections)
        if tags:
            click.echo(sep.join(tags), nl=not null)
        return

    setids = site.query(collections, discarded, outdated, path, tags, missing=missing)
    if setids:
        sep = '\x00' if null else '\n'
        click.echo(sep.join(setids), nl=not null)


@marvcli.command('run', short_help='Run nodes for DATASETS')
@click.option('--node', 'selected_nodes', multiple=True,
              help='Run individual nodes instead of all nodes used by detail and listing'
              ' Use --list-nodes for a list of known nodes. Beyond that any node can be run'
              ' by referencing it with package.module:node')
@click.option('--list-nodes', is_flag=True, help='List known nodes instead of running')
@click.option('--list-dependent', is_flag=True, help='List nodes depending on selected nodes')
@click.option('--deps/--no-deps', default=True, show_default=True,
              help='Run dependencies of requested nodes')
@click.option('--exclude', 'excluded_nodes', multiple=True,
              help='Exclude individual nodes instead of running all nodes used by detail and listing')
@click.option('-f', '--force/--no-force',
              help='Force run of nodes whose output is already available from store')
# TODO: force-deps is bad as it does not allow to resume
# better: discard nodes selectively or discard --deps node
# discarding does not delete, but simply registers the most recent generation to be empty
@click.option('--force-deps/--no-force-deps',
              help='Force run of dependencies whose output is already available from store')
@click.option('--force-dependent/--no-force-dependent',
              help='Force run of all nodes depending on selected nodes, directly or indirectly')
@click.option('--keep/--no-keep', help='Keep uncommitted streams for debugging')
@click.option('--keep-going/--no-keep-going',
              help='In case of an exception keep going with remaining datasets')
@click.option('--detail/--no-detail', 'update_detail',
              help='Update detail view from stored node output')
@click.option('--listing/--no-listing', 'update_listing',
              help='Update listing view from stored node output')
@click.option('--cachesize', default=50, show_default=True,
              help='Number of messages to keep in memory for each stream')
@click.option('--col', '--collection', 'collections', multiple=True,
              help='Run nodes for all datasets of selected collections, use "*" for all')
@click.argument('datasets', nargs=-1)
@click.pass_context
def marvcli_run(ctx, datasets, deps, excluded_nodes, force, force_dependent,
                force_deps, keep, keep_going, list_nodes,
                list_dependent, selected_nodes, update_detail,
                update_listing, cachesize, collections):
    """Run nodes for selected datasets.

    Datasets are specified by a list of set ids, or --collection
    <name>, use --collection=* to run for all collections. --node in
    conjunction with --collection=* will pick those collections for
    which the selected nodes are configured.

    Set ids may be abbreviated to any uniquely identifying
    prefix. Suffix a prefix by '+' to match multiple.

    """
    if collections and datasets:
        ctx.fail('--collection and DATASETS are mutually exclusive')

    if list_dependent and not selected_nodes:
        ctx.fail('--list-dependent needs at least one selected --node')

    if force_dependent and not selected_nodes:
        ctx.fail('--force-dependent needs at least one selected --node')

    if not any([datasets, collections, list_nodes]):
        click.echo(ctx.get_help())
        ctx.exit(1)

    deps = 'force' if force_deps else deps
    force = force_deps or force

    site = create_app().site

    if '*' in collections:
        if selected_nodes:
            collections = [k for k, v in site.collections.items()
                           if set(v.nodes).issuperset(selected_nodes)]
            if not collections:
                ctx.fail('No collections have all selected nodes')
        else:
            collections = None
    else:
        for col in collections:
            if col not in site.collections:
                ctx.fail('Unknown collection: {}'.format(col))

    if list_nodes:
        for col in (collections or sorted(site.collections.keys())):
            click.echo('{}:'.format(col))
            for name in sorted(site.collections[col].nodes):
                if name == 'dataset':
                    continue
                click.echo('    {}'.format(name))
        return

    if list_dependent:
        for col in (collections or sorted(site.collections.keys())):
            click.echo('{}:'.format(col))
            dependent = {x for name in selected_nodes
                         for x in site.collections[col].nodes[name].dependent}
            for name in sorted(x.name for x in dependent):
                click.echo('    {}'.format(name))
        return

    errors = []

    setids = [SetID(x) for x in parse_setids(datasets)]
    if not setids:
        query = db.session.query(Dataset.setid)\
                           .filter(Dataset.discarded.isnot(True))\
                           .filter(Dataset.status.op('&')(STATUS_MISSING) == 0)
        if collections is not None:
            query = query.filter(Dataset.collection.in_(collections))
        setids = (SetID(x[0]) for x in query)

    for setid in setids:
        if PDB:
            site.run(setid, selected_nodes, deps, force, keep,
                     force_dependent, update_detail, update_listing,
                     excluded_nodes, cachesize=cachesize)
        else:
            try:
                site.run(setid, selected_nodes, deps, force, keep,
                         force_dependent, update_detail, update_listing,
                         excluded_nodes, cachesize=cachesize)
            except UnknownNode as e:
                ctx.fail('Collection {} has no node {}'.format(*e.args))
            except NoResultFound:
                click.echo('ERROR: unknown {!r}'.format(setid), err=True)
                if not keep_going:
                    raise
            except RequestedMessageTooOld as e:
                ctx.fail((
                    "{} pulled {} message {} not being in memory anymore."
                    "\nSee https://ternaris.com/marv-robotics/docs/patterns.html#reduce-separately"
                ).format(e.args[0]._requestor.node.name, e.args[0].handle.node.name, e.args[1]))
            except BaseException as e:
                errors.append(setid)
                if isinstance(e, KeyboardInterrupt):
                    log.warn('KeyboardInterrupt: aborting')
                    raise
                elif isinstance(e, DirectoryAlreadyExists):
                    click.echo("""
ERROR: Directory for node run already exists:
{!r}
In case no other node run is in progress, this is a bug which you are kindly
asked to report, providing information regarding any previous, failed node runs.
""".format(e.args[0]), err=True)
                    if not keep_going:
                        ctx.abort()
                else:
                    log.error('Exception occured for dataset %s:', setid, exc_info=True)
                    log.error('Error occured for dataset %s: %s', setid, e)
                    if not keep_going:
                        ctx.exit(1)
    if errors:
        log.error('There were errors for %r', errors)


@marvcli.command('scan')
@click.option('-n', '--dry-run', is_flag=True)
def marvcli_scan(dry_run):
    """Scan for new and changed files"""
    create_app().site.scan(dry_run)


@marvcli.command('tag')
@click.option('--add', multiple=True, help='Tags to add')
@click.option('--rm', '--remove', multiple=True, help='Tags to remove')
@click.argument('datasets', nargs=-1)
@click.pass_context
def marvcli_tag(ctx, add, remove, datasets):
    """Add or remove tags to datasets"""
    if not any([add, remove]) or not datasets:
        click.echo(ctx.get_help())
        ctx.exit(1)

    app = create_app()
    setids = parse_setids(datasets)
    app.site.tag(setids, add, remove)


@marvcli.group('comment')
def marvcli_comment():
    """Add or remove comments"""


@marvcli_comment.command('add')
@click.option('-u', '--user', prompt=True, help='Commenting user')
@click.option('-m', '--message', required=True, help='Message for the comment')
@click.argument('datasets', nargs=-1)
def marvcli_comment_add(user, message, datasets):
    """Add comment as user for one or more datasets"""
    app = create_app()
    try:
        db.session.query(User).filter(User.name==user).one()
    except NoResultFound:
        click.echo("ERROR: No such user '{}'".format(user), err=True)
        sys.exit(1)
    ids = parse_setids(datasets, dbids=True)
    app.site.comment(user, message, ids)


@marvcli_comment.command('list')
@click.argument('datasets', nargs=-1)
def marvcli_comment_list(datasets):
    """Lists comments for datasets.

    Output: setid comment_id date time author message
    """
    app = create_app()
    ids = parse_setids(datasets, dbids=True)
    comments = db.session.query(Comment)\
                         .options(db.joinedload(Comment.dataset))\
                         .filter(Comment.dataset_id.in_(ids))
    for comment in sorted(comments, key=lambda x: (x.dataset._setid, x.id)):
        print(comment.dataset.setid, comment.id,
              datetime.datetime.fromtimestamp(int(comment.time_added / 1000)),
              comment.author, repr(comment.text))


SHOW_TEMPLATE = Template("""
{% for dataset in datasets %}
- name: {{ dataset.name }}
  collection: {{ dataset.collection }}
  setid: {{ dataset.setid }}
  files:
  {% for file in dataset.files %}
    - path: {{ file.path }}
      size: {{ file.size }}
  {% endfor %}

{% endfor %}
""".strip(), trim_blocks=True, lstrip_blocks=True)


@marvcli.command('show')
@click.argument('datasets', nargs=-1, required=True)
@click.pass_context
def marvcli_show(ctx, datasets):
    """Show information for one or more datasets in form of a yaml document.

    Set ids may be abbreviated as long as they are unique.
    \b
    Some examples
      marv show setid  # show one dataset
      marv query --col=* | xargs marv show   # show all datasets
    """
    app = create_app()
    ids = parse_setids(datasets, dbids=True)
    datasets = db.session.query(Dataset)\
                         .options(db.joinedload(Dataset.files))\
                         .filter(Dataset.id.in_(ids))\
                         .all()
    yamldoc = SHOW_TEMPLATE.render(datasets=datasets)
    print(yamldoc, end='')


@marvcli_comment.command('rm')
@click.argument('ids', nargs=-1)
def marvcli_comment_rm(ids):
    """Remove comments.

    Remove comments by id as given in second column of: marv comment list
    """
    app = create_app()
    db.session.query(Comment)\
              .filter(Comment.id.in_(ids))\
              .delete(synchronize_session=False)
    db.session.commit()


@marvcli.group('user')
def marvcli_user():
    """Manage user accounts"""


@marvcli_user.command('add')
@click.option('--password', help='Password will be prompted')
@click.argument('username')
@click.pass_context
def marvcli_user_add(ctx, username, password):
    """Add a user"""
    if not re.match(r'[0-9a-zA-Z\-_\.@+]+$', username):
        click.echo('Invalid username: {}'.format(username), err=True)
        click.echo('Must only contain ASCII letters, numbers, dash, underscore and dot',
                   err=True)
        sys.exit(1)
    if password is None:
        password = click.prompt('Password', hide_input=True, confirmation_prompt=True)
    app = create_app()
    try:
        app.um.user_add(username, password, 'marv', '')
    except ValueError as e:
        click.echo('Error: {}'.format(e.args[0], err=True))
        sys.exit(1)


@marvcli_user.command('list')
def marvcli_user_list():
    """List existing users"""
    app = create_app()
    query = db.session.query(User).options(db.joinedload(User.groups))\
                                  .order_by(User.name)
    users = [(user.name, ', '.join(sorted(x.name for x in user.groups)))
             for user in query]
    users = [('User', 'Groups')] + users
    uwidth, gwidth = reduce(lambda (xm, ym), (x, y): (max(x, xm), max(y, ym)),
                            ((len(x), len(y)) for x, y in users))
    fmt = '{:%ds} | {}' % uwidth
    for x, y in users:
        click.echo(fmt.format(x, y))


@marvcli_user.command('pw')
@click.option('--password', prompt=True, hide_input=True, confirmation_prompt=True,
              help='Password will be prompted')
@click.argument('username')
@click.pass_context
def marvcli_user_pw(ctx, username, password):
    """Change password"""
    app = create_app()
    try:
        app.um.user_pw(username, password)
    except ValueError as e:
        ctx.fail(e.args[0])


@marvcli_user.command('rm')
@click.argument('username')
@click.pass_context
def marvcli_user_rm(ctx, username):
    """Remove a user"""
    app = create_app()
    try:
        app.um.user_rm(username)
    except ValueError as e:
        ctx.fail(e.args[0])


@marvcli.group('group')
def marvcli_group():
    """Manage user groups"""


@marvcli_group.command('add')
@click.argument('groupname')
@click.pass_context
def marvcli_group_add(ctx, groupname):
    """Add a group"""
    if not re.match(r'[0-9a-zA-Z\-_\.@+]+$', groupname):
        click.echo('Invalid groupname: {}'.format(groupname), err=True)
        click.echo('Must only contain ASCII letters, numbers, dash, underscore and dot',
                   err=True)
        sys.exit(1)
    app = create_app()
    try:
        app.um.group_add(groupname)
    except ValueError as e:
        ctx.fail(e.args[0])


@marvcli_group.command('list')
def marvcli_group_list():
    """List existing groups"""
    app = create_app()
    query = db.session.query(Group.name).order_by(Group.name)
    for name in [x[0] for x in query]:
        click.echo(name)


@marvcli_group.command('adduser')
@click.argument('username')
@click.argument('groupname')
@click.pass_context
def marvcli_group_adduser(ctx, groupname, username):
    """Add an user to a group"""
    app = create_app()
    try:
        app.um.group_adduser(groupname, username)
    except ValueError as e:
        ctx.fail(e.args[0])


@marvcli_group.command('rmuser')
@click.argument('username')
@click.argument('groupname')
@click.pass_context
def marvcli_group_rmuser(ctx, groupname, username):
    """Remove an user from a group"""
    app = create_app()
    try:
        app.um.group_rmuser(groupname, username)
    except ValueError as e:
        ctx.fail(e.args[0])


@marvcli_group.command('rm')
@click.argument('groupname')
@click.pass_context
def marvcli_group_rm(ctx, groupname):
    """Remove a group"""
    app = create_app()
    try:
        app.um.group_rm(groupname)
    except ValueError as e:
        ctx.fail(e.args[0])



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import json
from json import dumps as jsondumps

import flask
import pendulum
from flask import current_app, request
from sqlalchemy.sql import select

from marv.collection import Filter, UnknownOperator
from marv.model import STATUS, Tag, db
from marv.utils import parse_datetime, parse_filesize, parse_timedelta
from .tooling import APIEndpoint, api_endpoint as marv_api_endpoint


ALIGN = {
    'filesize': 'right',
    'icon': 'center',
    'int': 'right',
    'timedelta': 'right',
}


FILTER_PARSER = {
    'datetime': lambda x: int(
        (parse_datetime(x) - parse_datetime('1970-01-01T00:00:00+00:00'))
        .total_seconds() * 1000
    ),
    'filesize': parse_filesize,
    'float': float,
    'int': int,
    'string': str,
    'string[]': str,
    'subset': lambda x: x,
    'timedelta': parse_timedelta,
    'words': lambda x: x.split(),
}


VALUE_TYPE_MAP = {
    'string[]': 'string',
    'words': 'string',
}


TIMEZONE = pendulum.local_timezone().name


# Order corresponds to marv.model.STATUS OrderedDict
STATUS_ICON = ['fire', 'eye-close', 'warning-sign', 'time']
STATUS_JSON = [jsondumps({'icon': STATUS_ICON[i], 'title': x},
                         separators=(',', ':'))
               for i, x in enumerate(STATUS.values())]
# TODO: reconsider in case we get a couple of more states
STATUS_STRS = {
    bitmask: ','.join(filter(None,
                             [STATUS_JSON[i] if bitmask & 2**i else None
                              for i in range(len(STATUS_JSON))]))
    for bitmask in range(2**(len(STATUS_JSON)))
}


class FilterParseError(Exception):
    pass


def parse_filters(specs, filters):
    filters = json.loads(filters)
    filters = [Filter(k, FILTER_PARSER[specs[k].value_type](v['val']), v['op'],
                      specs[k].value_type)
               for k, v in filters.items()]
    return filters


@marv_api_endpoint('/meta', force_acl=['__unauthenticated__', '__authenticated__'])
def meta():
    groups = request.user_groups

    collection_acl = current_app.acl['collection']
    if '__unauthenticated__' not in collection_acl and not request.username:
        return flask.jsonify({
            'realms': current_app.site.config.marv.oauth.keys(),
        })

    # TODO: avoid clashes, forbid routes with same name / different name generation
    acl = sorted([
        name.split('.')[-1]
        for name, view in current_app.view_functions.viewitems()
        if isinstance(view, APIEndpoint)
        if view.acl.intersection(groups)
    ])

    # TODO: id probably numeric, title = name or title
    collections = current_app.site.collections
    collection_id = collections.default_id
    resp = flask.jsonify({
        'collections': {
            'default': collection_id,
            'items': [{'id': x, 'title': x} for x in collections.keys()],
        },
        'acl': acl,
        'realms': current_app.site.config.marv.oauth.keys(),
        'timezone': TIMEZONE,
    })
    resp.headers['Cache-Control'] = 'no-cache'
    return resp


@marv_api_endpoint('/collection', defaults={'collection_id': None})
@marv_api_endpoint('/collection/<collection_id>')
def collection(collection_id):
    collections = current_app.site.collections
    collection_id = collections.default_id if collection_id is None else collection_id
    try:
        collection = collections[collection_id]
    except KeyError:
        flask.abort(404)

    try:
        filters = parse_filters(collection.filter_specs,
                                request.args.get('filter', '{}'))
    except (KeyError, ValueError):
        current_app.logger.warn('Bad request', exc_info=True)
        return flask.abort(400)

    try:
        stmt = collection.filtered_listing(filters)
    except UnknownOperator:
        current_app.logger.warn('Bad request', exc_info=True)
        flask.abort(400)

    fspecs = collection.filter_specs

    all_known = {
        name: [x[0] for x in
               db.session.execute(select([rel.value])
                                  .order_by(rel.value)).fetchall()]
        for name, rel in collection.model.relations.items()
        if {'any', 'all'}.intersection(fspecs[name].operators)
    }
    all_known['status'] = STATUS.keys()
    all_known['tags'] = [x[0] for x in
                         db.session.execute(select([Tag.value])
                                            .where(Tag.collection == collection.name)
                                            .order_by(Tag.value)).fetchall()]
    filters = [{'key': x.name,
                'constraints': all_known.get(x.name),
                'title': x.title,
                'operators': x.operators,
                'value_type': VALUE_TYPE_MAP.get(x.value_type, x.value_type)}
               for x in fspecs.values()]

    rows = db.session.execute(stmt).fetchall()
    rowdata = ',\n'.join([x.replace('["#TAGS#"]', tags if tags != '[null]' else '[]')
                          .replace('"#TAGS#"', tags[1:-1] if tags != '[null]' else '')
                          .replace('[,', '[')
                          .replace('"#STATUS#"', STATUS_STRS[status])
                          for x, status, tags in rows])
    dct = {
        'all_known': all_known,
        'compare': bool(collection.compare),
        'filters': {'title': 'Filter',
                    'widget': {'type': 'filter', 'filters': filters}},
        'summary_config': {'title': 'Summary', 'items': collection.summary_items},
        'listing': {'title': 'Data sets ({} found)'.format(len(rows)),
                    'widget': {
                        'data': {
                            'columns': [{
                                'align': ALIGN.get(x.formatter, 'left'),
                                'formatter': x.formatter,
                                'sortkey': 'title' if x.formatter == 'route' else None,
                                'id': x.name,
                                'title': x.heading,
                                'list': x.islist,
                            } for x in collection.listing_columns],
                            'rows': ['#ROWS#'],
                            'sortcolumn': collection.sortcolumn,
                            'sortorder': collection.sortorder},
                        'type': 'table',
                    }},
    }
    indent = None
    separators = (',', ':')
    if current_app.config['JSONIFY_PRETTYPRINT_REGULAR'] or current_app.debug:
        indent = 2
        separators = (', ', ': ')
    jsondata = jsondumps(dct, indent=indent, separators=separators, sort_keys=True)
    jsondata = jsondata.replace('"#ROWS#"', rowdata)
    resp = current_app.response_class(
        (jsondata, '\n'), mimetype=current_app.config['JSONIFY_MIMETYPE'])
    resp.headers['Cache-Control'] = 'no-cache'
    return resp



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import unittest

from ..setid import SetID


class TestCase(unittest.TestCase):
    def test_setid(self):
        self.assertNotEqual(SetID(42), SetID(17))
        self.assertEqual(SetID(42), SetID('fiaaaaaaaaaaaaaaaaaaaaaaaa'))
        self.assertEqual(repr(SetID(42)), "SetID('fiaaaaaaaaaaaaaaaaaaaaaaaa')")
        self.assertEqual(str(SetID(42)), 'fiaaaaaaaaaaaaaaaaaaaaaaaa')
        self.assertEqual(unicode(SetID(42)), u'fiaaaaaaaaaaaaaaaaaaaaaaaa')
        setid = SetID(42)
        self.assertEqual(setid.lohi, (setid.lo, setid.hi))
        self.assertEqual(setid.lohi, (42, 0))
        self.assertEqual(int(setid), 42)
        self.assertEqual(long(setid), 42)
        self.assertEqual(SetID(42, 0), SetID(42))
        self.assertEqual(SetID(42, 1), SetID((1 << 64) + 42))



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import os
import shutil
import tempfile
import unittest
from contextlib import contextmanager


@contextmanager
def chdir(d):
    """Change working directory - NOT THREAD SAFE"""
    cwd = os.getcwd()
    os.chdir(d)
    try:
        yield d
    finally:
        os.chdir(cwd)


def make_scanroot(scanroot, names):
    if not os.path.exists(scanroot):
        os.makedirs(scanroot)
    for name in names:
        with open(os.path.join(scanroot, name), 'w') as f:
            f.write(name)


@contextmanager
def temporary_directory(keep=None):
    """Create, change into, and cleanup temporary directory"""
    tmpdir = tempfile.mkdtemp()
    with chdir(tmpdir):
        try:
            yield tmpdir
        finally:
            if not keep:
                shutil.rmtree(tmpdir)


def decode(data, encoding='utf-8'):
    if isinstance(data, str):
        data = data.decode(encoding)
    elif isinstance(data, dict):
        data = {decode(k): decode(v) for k, v in data.items()}
    elif isinstance(data, list):
        data = [decode(x) for x in data]
    elif isinstance(data, tuple):
        data = tuple(decode(x) for x in data)
    return data


class TestCase(unittest.TestCase):
    """Basic marv test case"""
    KEEP_TEST_DIR = None
    test_dir = None

    def run(self, result=None):
        with temporary_directory(self.KEEP_TEST_DIR) as tmpdir:
            self.test_dir = tmpdir
            super(TestCase, self).run(result)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import base64
import bcrypt
import fcntl
import json
import os
import time

import flask
import jwt
from flask import current_app
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from marv import utils
from marv.model import User, Group, db
from .tooling import api_group as marv_api_group


# TODO: switch to idempotent OR IGNORE (like tag.py)
# TODO: move (part) of this to site
class UserManager(object):
    def __init__(self):
        pass

    def authenticate(self, username, password):
        if not username or not password:
            return False
        try:
            user = db.session.query(User).filter_by(name=username, realm='marv').one()
        except NoResultFound:
            return False
        hashed = user.password.encode('utf-8')
        return bcrypt.hashpw(password, hashed) == hashed

    def user_add(self, name, password, realm, realmuid, given_name=None, family_name=None,
                 email=None, time_created=None, time_updated=None, _restore=None):
        try:
            if not _restore:
                password = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt())
            now = int(utils.now())
            if not time_created:
                time_created = now
            if not time_updated:
                time_updated = now
            user = User(name=name, password=password, realm=realm, given_name=given_name,
                        family_name=family_name, email=email, realmuid=realmuid,
                        time_created=time_created, time_updated=time_updated)
            db.session.add(user)
            db.session.commit()
        except IntegrityError:
            raise ValueError('User {} exists already'.format(name))

    def user_rm(self, username):
        try:
            user = db.session.query(User).filter_by(name=username).one()
            db.session.delete(user)
            db.session.commit()
        except NoResultFound:
            raise ValueError('User {} does not exist'.format(username))

    def user_pw(self, username, password):
        try:
            user = db.session.query(User).filter_by(name=username).one()
        except NoResultFound:
            raise ValueError('User {} does not exist'.format(username))

        user.password = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt())
        user.time_updated = int(utils.now())
        db.session.commit()

    def group_add(self, groupname):
        try:
            group = Group(name=groupname)
            db.session.add(group)
            db.session.commit()
        except IntegrityError:
            raise ValueError('Group {} exists already'.format(groupname))

    def group_rm(self, groupname):
        try:
            group = db.session.query(Group).filter_by(name=groupname).one()
            db.session.delete(group)
            db.session.commit()
        except NoResultFound:
            raise ValueError('Group {} does not exist'.format(groupname))

    def group_adduser(self, groupname, username):
        try:
            group = db.session.query(Group).filter_by(name=groupname).one()
        except NoResultFound:
            raise ValueError('Group {} does not exist'.format(groupname))
        try:
            user = db.session.query(User).filter_by(name=username).one()
        except NoResultFound:
            raise ValueError('User {} does not exist'.format(username))
        group.users.append(user)
        db.session.commit()

    def group_rmuser(self, groupname, username):
        try:
            group = db.session.query(Group).filter_by(name=groupname).one()
        except NoResultFound:
            raise ValueError('Group {} does not exist'.format(groupname))
        try:
            user = db.session.query(User).filter_by(name=username).one()
        except NoResultFound:
            raise ValueError('User {} does not exist'.format(username))
        if user in group.users:
            group.users.remove(user)
        db.session.commit()

    def check_authorization(self, acl, authorization):
        username = None
        groups = {'__unauthenticated__'}
        if authorization:
            try:
                token = authorization.replace('Bearer ', '')
                session = jwt.decode(token, current_app.config['SECRET_KEY'])
                user = db.session.query(User).filter_by(name=session['sub']).one()
            except:
                flask.abort(401)
            if user.time_updated > session['iat']:
                flask.abort(401)
            username = user.name
            groups = set([g.name for g in user.groups])
            groups.add('__authenticated__')

        elif '__unauthenticated__' not in acl:
            flask.abort(401)

        if acl and not len(acl.intersection(groups)):
            flask.abort(403)

        flask.request.username = username
        flask.request.user_groups = groups


@marv_api_group()
def auth(app):
    assert not hasattr(app, 'um')  # TODO: correct place for extension
    app.um = UserManager()


def generate_token(username):
    now = int(time.time())
    return jwt.encode({
        'exp': now + 2419200, # 4 weeks expiration
        'iat': now,
        'sub': username,
    }, current_app.config['SECRET_KEY'])


@auth.endpoint('/auth', methods=['POST'], force_acl=['__unauthenticated__'])
def auth_post():
    req = flask.request.get_json()
    if not req:
        flask.abort(400)
    username = req.get('username', '')
    password = req.get('password', '').encode('utf-8')

    if not current_app.um.authenticate(username, password):
        return flask.abort(422)

    return flask.jsonify({'access_token': generate_token(username)})



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import unittest


class TestCase(unittest.TestCase):
    def test_(self):
        from . import pythonic_capnp

        # class Pythonic(Base):
        #     """foo"""
        #     schema = pythonic_capnp.Pythonic

        import capnp
        p = pythonic_capnp.Pythonic.new_message()
        r = p.as_reader()



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

"""Marv config parsing."""
from __future__ import absolute_import, division, print_function

import os
import sys
from inspect import getmembers
from collections import Mapping
from functools import partial
from logging import getLogger

from configparser import ConfigParser

from .utils import find_obj

log = getLogger(__name__)


class ConfigError(Exception):
    def __str__(self):
        try:
            section, key, message = self.args
        except ValueError:
            section, key = self.args
            message = 'missing'
        return '{} [{}] {}: {}'.format(section.filename, section.name, key, message)


def make_funcs(dataset, setdir, store):
    """Functions available for listing columns and filters."""
    return {
        'cat': lambda *lists: [x for lst in lists for x in lst],
        'comments': lambda: None,
        'detail_route': detail_route,
        'format': lambda fmt, *args: fmt.format(*args),
        'get': partial(getnode, dataset, setdir, store),
        'join': lambda sep, *args: sep.join([x for x in args if x]),
        'len': len,
        'link': (lambda href, title, target=None:
                 {'href': href or "",
                  'title': title or "",
                  'target': '_blank' if target is None else target}),
        'list': lambda *x: filter(None, list(x)),
        'max': max,
        'min': min,
        'status': lambda: ['#STATUS#'],
        'sum': sum,
        'tags': lambda: ['#TAGS#'],
        'trace': print_trace,
    }


def make_summary_funcs(rows, ids):
    """Functions available for listing summary fields."""
    return {
        'len': len,
        'list': lambda *x: filter(None, list(x)),
        'max': max,
        'min': min,
        'rows': partial(summary_rows, rows, ids),
        'sum': sum,
        'trace': print_trace
    }


def summary_rows(rows, ids, id=None, default=None):
    if not id:
        return rows
    idx = ids.index(id)
    return [default if value is None else value
            for value in (x['values'][idx] for x in rows)]


def calltree(functree, funcs):
    name, args = functree
    func = funcs[name]
    args = [calltree(x, funcs) if type(x) == tuple else x for x in args]
    return func(*args)


def getdeps(functree, deps=None):
    deps = set() if deps is None else deps
    name, args = functree
    if name == 'get':
        deps.add(getnode(None, None, None, args[0], _name_only=True))
    else:
        for arg in args:
            if type(arg) == tuple:
                getdeps(arg, deps)
    return deps


def detail_route(setid, name=None):
    return {
        'route': 'detail',
        'id': str(setid),
        'title': name if name else str(setid),
    }


def doget(value, name, lookup):
    value = getattr(value, name, None)
    if value is None:
        return None
    if lookup is not None:
        try:
            return value[lookup]
        except IndexError:
            return None
    return value


def getnode(dataset, setdir, store, objpath, default=None, _name_only=False):
    try:
        nodename, rest = objpath.split('.', 1)
    except ValueError:
        nodename, rest = objpath, None
    nodename, lookup = parse_lookup(nodename, 0)

    if _name_only:
        return nodename

    try:
        node = store.nodes[nodename]
    except KeyError:
        log.critical('Config error: unknown node %s', nodename)
        sys.exit(1)

    if node.name == 'dataset':
        msgs = node.load(setdir, dataset)
    else:
        try:
            msgs = store.load(setdir, node)
        except IOError:  # TODO: Dedicated exception
            return default

    try:
        value = msgs[lookup]
    except IndexError:
        return default

    # traverse the rest
    while rest:
        try:
            name, rest = rest.split('.', 1)
        except ValueError:
            name, rest = rest, None
        name, lookup = parse_lookup(name)
        if isinstance(value, list):
            value = [doget(x, name, lookup) for x in value]
        else:
            value = doget(value, name, lookup)
    return value


def parse_lookup(name, default=None):
    if name.endswith(']'):
        name, lookup = name[:-1].split('[')
        if ':' in lookup:
            lookup = slice(*[int(x) if x else None for x in lookup.split(':')])
        else:
            lookup = int(lookup)
    else:
        lookup = default
    return name, lookup


def print_trace(args):
    print("TRACE", args)
    return args


def parse_function(s, pos=0):
    assert s[pos] == '(', s
    start = pos + 1
    end = len(s) - 1
    try:
        pos = s.index(' ', start)
    except ValueError:
        pos = end
    name = s[start:pos]
    pos += 1
    args = []
    functree = (name, args)
    while s[pos] != ')' and pos < end:
        if s[pos] == '(':
            #print('FUNC', pos, s[pos:])
            func, pos = parse_function(s, pos)
            args.append(func)
        elif s[pos:pos+2] == '[]':
            args.append([])
            pos += 2
        elif s[pos] == '"':
            #print('STR"', pos, s[pos:])
            start = pos + 1
            pos = s.index('"', start)
            args.append(s[start:pos])
            pos += 1
        elif s[pos] == "'":
            #print("STR'", pos, s[pos:])
            start = pos + 1
            pos = s.index("'", start)
            args.append(s[start:pos])
            pos += 1
        elif s[pos] == '0':
            args.append(0)
            pos += 1
        elif s[pos] == ' ':
            pos += 1
        else:
            raise RuntimeError(pos, s[pos:])
    return functree, pos + 1


class Section(Mapping):
    def __init__(self, name, dct, filename, defaults=None, schema=None):
        self.name = name
        self._configdir = os.path.dirname(filename)
        self.filename = filename
        self._dct = dct
        self._defaults = defaults or {}
        self._schema = schema or {}

    def __dir__(self):
        return list(self._dct.viewkeys() | self.__dict__.viewkeys() |
                    {x[0] for x in getmembers(type(self))})

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, key):
        try:
            value = self._dct[key]
        except KeyError:
            value = self._defaults[key]
        if value is None:
            return None
        value = value.strip()
        value_type = self._schema.get(key)
        if value_type is not None:
            value = self._parse(value, value_type)
        return value

    def __iter__(self):
        return iter(list(self._dct.viewkeys() | self._defaults.viewkeys()))

    def __len__(self):
        return len(self._dct.viewkeys() | self._defaults.viewkeys())

    def _parse(self, value, value_type):
        if value_type == 'function':
            # TODO: lookup node to fail early
            func, pos = parse_function(value)
            assert pos == len(value)
            return func
        elif value_type == 'find_obj':
            return find_obj(value)
        elif value_type == 'lines':
            return [x.strip() for x in value.splitlines()]
        elif value_type == 'nodes':
            return [find_obj(x.strip()) for x in value.splitlines()]
        elif value_type == 'path':
            return self._abspath(value)
        elif value_type == 'path_lines':
            return [self._abspath(x.strip()) for x in value.splitlines()]
        elif value_type == 'pipe_separated_list':
            return [x.strip() for x in value.split('|')]
        elif value_type == 'space_separated_list':
            return value.split()
        else:
            raise ValueError('Unknown value_type %r' % value_type)

    def _abspath(self, value):
        if os.path.isabs(value):
            return value
        else:
            return os.path.realpath(os.path.join(self._configdir, value))


class Config(Mapping):
    def __init__(self, filename, sections):
        assert os.path.isabs(filename), filename
        self.filename = filename
        self._dct = sections

    @classmethod
    def from_file(cls, filename, defaults=None, required=None, schema=None):
        assert os.path.isabs(filename), filename
        defaults = defaults or {}
        schema = schema or {}

        # TODO: replace by own parser supporting line numbers
        parser = ConfigParser()
        with open(filename) as f:
            parser.read_file(f)

        sections = {
            k: Section(name=k, dct=v, filename=filename,
                       defaults=defaults.get(k) or defaults.get(k.split(None, 1)[0]),
                       schema=(schema.get(k) or schema.get(k.split(None, 1)[0])))
            for k, v in parser._sections.items()
        }
        if required:
            for name, section in sections.items():
                reqs = required.get(name) or required.get(name.split(None, 1)[0], [])
                for req in reqs:
                    if req not in section:
                        raise ConfigError(section, req)
        return cls(filename, sections)

    def __dir__(self):
        return list(self._dct.viewkeys() | self.__dict__.viewkeys() |
                    {x[0] for x in getmembers(type(self))})

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, key):
        return self._dct[key]

    def __iter__(self):
        return iter(self._dct)

    def __len__(self):
        return len(self._dct)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import functools
import os
from collections import OrderedDict, deque
from logging import getLogger
from pprint import pformat

from marv_node.setid import SetID
from marv_nodes import dataset as dataset_node
from marv_store.streams import ReadStream

from .driver import Driver
from .event import DefaultOrderedDict
from .io import NEXT, PAUSED, RESUME, THEEND
from .io import Task, MsgRequest
from .stream import Handle, Msg, Stream, VolatileStream


class UnmetDependency(Exception):
    pass


PPINFO = os.environ.get('PPINFO')
MARV_RUN_LOGBREAK = os.environ.get('MARV_RUN_LOGBREAK')
RAISE_IF_UNFINISHED = False


def run_nodes(dataset, nodes, store, persistent=None, force=None,
              deps=None, cachesize=None):

    if cachesize is not None:
        import marv_node
        marv_node.stream.CACHESIZE
        marv_node.stream.CACHESIZE = cachesize

    queue = []
    ret = run_nodes_async(dataset=dataset, nodes=nodes, store=store,
                          queue=queue, persistent=persistent,
                          force=force, deps=deps)
    if ret is None:
        return False

    streams, loop, process_task, after = ret

    # initialize nodes and get initial set of promises
    done = False
    send_queue_empty = False
    while not done:
        # comment while for synchronous
        # while not done and not send_queue_empty:
        #     # additional random break
        #     done, send_queue_empty = process_task(None, None)
        # # random pick from queue
        current, task = queue.pop(0) if queue else (None, None)
        done, send_queue_empty = process_task(current, task)
    after()

    return streams


def run_nodes_async(dataset, nodes, store, queue, persistent=None,
                    force=None, deps=None):
    deps = True if deps is None else deps
    force = False if force is None else force
    assert len(nodes) == len(set(nodes)), nodes
    assert deps in (False, True, 'force'), deps
    assert not any(x.name == 'dataset' for x in nodes)
    persistent = {node: name for name, node in (persistent or {}).items()}
    setid = dataset.setid
    assert isinstance(setid, SetID), setid

    log = getLogger('marv.run')
    logdebug = log.debug
    lognoisy = log.noisy
    logverbose = log.verbose
    loginfo = log.info
    logerror = log.error

    def getname(node):
        return persistent.get(node, node.abbrev)

    drivers = OrderedDict()
    send_queue = deque()

    done = OrderedDict()     # drivers that are finished
    serving = OrderedDict()  # stream.handle -> driver
    streams = OrderedDict()  # stream.handle -> stream
    suspended = []  # suspended drivers
    # {(stream.handle, msg.idx): [driver1, ...]}
    waiting = DefaultOrderedDict(list)
    loggers = OrderedDict()

    def start_driver(driver, force=False):
        assert driver.key not in drivers, driver
        drivers[driver.key] = driver
        send_queue.append((driver, driver.start()))
        store_name = persistent.get(driver.node)

        class LoggerProxy(object):
            def __init__(self, name, prefix):
                self.logger = getLogger(name)
                self.prefix = prefix

            def __getattr__(self, name):
                meth = getattr(self.logger, name)
                @functools.wraps(meth)
                def newmeth(msg, *args, **kw):
                    msg = self.prefix + msg
                    if msg % args == MARV_RUN_LOGBREAK:
                        import ipdb
                        ipdb.set_trace()
                    return meth(msg, *args, **kw)
                return newmeth

        logprefix = '.'.join([setid.abbrev, driver.node.abbrev, driver.name]) + ' '
        logger = LoggerProxy('marv.run', logprefix)
        loggers[driver] = logger
        logmeth = getattr(logger, 'info' if store_name else 'verbose')
        logmeth('%sstarted%s',
                '({}) '.format(store_name) if store_name else '',
                ' with force' if force else '')

    logverbose('evaluating %s %s', setid.abbrev, ' '.join(sorted(getname(x) for x in nodes)))
    for node in sorted(nodes, key=getname):
        handle = Handle(setid=setid, node=node, name='default')
        if handle in store:
            if not force:
                logverbose('skipping stored %s', getname(node))
                continue
        stream = store.create_stream(handle) if node in persistent \
                 else VolatileStream(handle)
        driver = Driver(stream)
        start_driver(driver, force and handle in store)

    if not drivers:
        logverbose('all satisfied.')
        return None

    # TODO: reconsider what pulling means and if needed
    # Drivers for nodes passed to run_nodes and their forks are the
    # ones pulling. Warn if after that there are persistent nodes
    # unfinished and continue until they are finished.
    pulling = OrderedDict((x, None) for x in drivers.values())

    dataset_handle = Handle(setid, dataset_node, 'default')
    streams[dataset_handle] = ReadStream(dataset_handle, None, None,
                                         dataset_node.load(None, dataset))

    def queue_front(driver, send):
        assert driver not in {driver for lst in waiting.values()
                              for driver in lst}
        if driver in suspended:
            suspended.remove(driver)
        assert driver not in suspended
        send_queue.appendleft((driver, send))
        loggers[driver].debug('QUEUE_FRONT %r', send)
        if PPINFO:
            logdebug("state %s", ppinfo())

    def queue_back(driver, send):
        assert driver not in {driver for lst in waiting.values()
                              for driver in lst}
        if driver in suspended:
            suspended.remove(driver)
        assert driver not in suspended
        send_queue.append((driver, send))
        loggers[driver].debug('QUEUE %r', send)
        if PPINFO:
            logdebug("state %s", ppinfo())

    def suspend(driver, _):
        assert driver not in {driver for lst in waiting.values()
                              for driver in lst}
        assert driver not in suspended
        assert driver not in pulling
        suspended.append(driver)
        loggers[driver].debug('SUSPEND')

    def ppinfo():
        lines = ['']
        lines.extend('  ' + x for x in pformat([
            ('drivers', list(drivers.values())),
            ('done', list(done)),
            ('queue', list(queue)),
            ('send_queue', ['{!r} {!r}'.format(x, y) for x, y in send_queue]),
            ('suspended', list(suspended)),
            ('pulling', list(pulling)),
            ('waiting', sorted(['{!r} {!r} {}'.format(driver, handle, idx)
                                for (handle, idx), lst in waiting.items()
                                for driver in lst])),
            ('streams', [(repr(v), v.info()) for k, v in streams.items()]),
        ]).split('\n'))
        return '\n'.join(lines)

    def ppwait():
        return pformat([
            '{} {} {}'.format(driver.key_abbrev, handle.key_abbrev, idx)
            for (handle, idx), lst in waiting.items()
            for driver in lst
        ])
    def pppinfo():
        print(ppinfo())


    class Counter:
        msgnum = 0

    def loop():
        pppinfo  # make available in context for debugging; somebody invented classes...
        if not send_queue:
            # simply wake up all suspended as long as there are
            # unfinished pulling drivers
            if any(driver for lst in waiting.values()
                   for driver in lst
                   if driver in pulling):
                for driver in suspended:
                    queue_back(driver, RESUME)

        if not send_queue:
            return Counter.msgnum == 0, True

        current, send = send_queue.popleft()
        assert current not in {driver for lst in waiting.values()
                               for driver in lst}
        assert current not in suspended
        try:
            if isinstance(send, Msg):
                msg = send
                loggers[current].noisy('<- %s %s',
                                       'HANDLE' if msg.idx == -1 else msg.idx,
                                       msg.handle.key_abbrev)
            else:
                loggers[current].debug('<- %r', send)
            promise = current.send(send)
            assert isinstance(promise, Task), promise
            queue.append((current, promise))
            Counter.msgnum += 1
            #loggers[current].debug('recv %r', promise)
        except StopIteration:
            methname = 'info' if current.node in persistent else 'verbose'
            logmeth = getattr(loggers[current], methname)
            logmeth('finished')
            assert current not in done, current
            done[current] = None

        return False, False

    def process_task(current, task):
        if not (current is None and task is None):
            Counter.msgnum -= 1

        if current is None and task is None:
            return loop()

        logger = loggers[current]

        if task is PAUSED:
            meth = (queue_back if current in pulling else suspend)
            meth(current, RESUME)
            return loop()

        elif isinstance(task, Driver):
            driver = task
            logger.noisy('FORK %s', driver.name)
            start_driver(driver)
            if current in pulling:
                pulling[driver] = None
            queue_back(current, NEXT)
            return loop()

        elif isinstance(task, Stream):
            stream = task
            # TODO: in case of a restart this is ok, see also comment
            # below near reqdriver
            #
            #assert stream.handle not in streams, stream.handle
            #assert stream.handle not in serving
            streams[stream.handle] = stream
            serving[stream.handle] = current
            logger.noisy('ADDSTREAM %s', stream.name)
            queue_front(current, NEXT)
            return loop()

        elif isinstance(task, Msg):
            msg = task
            assert msg.handle.node is current.node
            assert msg.handle.setid == current.setid
            if msg.idx == -1:
                logger.noisy('PUBHANDLE %s', msg.handle.name)
            else:
                logger.noisy('-> %d %s%s',
                             msg.idx,
                             msg.handle.name,
                             ' DONE' if msg.data == THEEND else '')
            stream = streams[msg.handle]
            stream.add_msg(msg)
            waitees = waiting.pop((msg.handle, msg.idx), [])
            for waitee in waitees:
                queue_back(waitee, msg)
            if msg.idx == -1 and stream.parent is not None:
                assert isinstance(msg.data, Handle), msg.data
                msg = stream.parent.handle.msg(msg.data)
                stream.parent.add_msg(msg)
                waitees = waiting.pop((msg.handle, msg.idx), [])
                for waitee in waitees:
                    queue_back(waitee, msg)
            meth = (queue_back if current in pulling else suspend)
            meth(current, NEXT)
            return loop()

        elif isinstance(task, MsgRequest):
            req, (handle, idx) = task, task

            if idx == -1:
                logger.noisy('REQHANDLE %s', handle.key_abbrev)

            # Serve the request
            stream = streams.get(handle)
            forced = False
            if stream is None:
                try:
                    stream = store[handle]
                    if deps == 'force' and handle.node.name != 'dataset' or \
                       force and handle.node in nodes:
                        forced = True
                        stream = None
                    else:
                        logverbose('%r from store', stream)
                        streams[stream.handle] = stream
                except KeyError:
                    if not deps:
                        logerror('%r not in store, running explicitly disabled', handle)
                        raise UnmetDependency()

            try:
                send = stream.get_msg(req) if stream is not None else None
            except IndexError:
                logverbose("state %s", ppinfo())
                logerror('Failed to get %r for %r', req, current)
                raise

            if send is not None:
                queue_back(current, send)
                return loop()

            # Request has to wait
            waitees = waiting[(handle, idx)]
            assert current not in waitees
            waitees.append(current)
            logger.debug('AWAIT %s %s',
                         'HANDLE' if req.idx == -1 else req.idx,
                         req.handle.key_abbrev)

            if handle in serving:
                return loop()

            # Get driver or instantiate it (and request substream)
            node, setid = handle.node, handle.setid
            driver_key = (setid, node, 'default')
            reqdriver = drivers.get(driver_key)
            handles = []
            if reqdriver and handle.name != 'default':
                assert reqdriver.node.group == 'ondemand'
                if not reqdriver.stream_creation:
                    loggers[reqdriver].verbose('restarting')
                    handles = list(reqdriver._requested_streams)
                    reqdriver.destroy()
                    del drivers[reqdriver.key]
                    if reqdriver in done:
                        del done[reqdriver]
                    if reqdriver in suspended:
                        suspended.remove(reqdriver)
                    queue = list(send_queue)
                    send_queue.clear()
                    send_queue.extend(x for x in queue if x[0] != reqdriver)
                    # TODO: If we delete other nodes might request things that don't exist
                    # For now the streams are simply replaced if a
                    # (restarted) driver returns the stream again
                    #
                    # for h, d in serving.items():
                    #     if d == reqdriver:
                    #         del serving[h]
                    #         del streams[h]
                    reqdriver = None
            if reqdriver is None:
                driver_handle = Handle(*driver_key)
                driver_stream = \
                    store.create_stream(driver_handle) if node in persistent else \
                    VolatileStream(driver_handle)
                reqdriver = Driver(driver_stream)
                assert reqdriver.key == driver_key
                start_driver(reqdriver, forced)
                if node in persistent:
                    pulling[reqdriver] = None
            if handle.name != 'default':
                handles.append(handle)
                reqdriver.add_stream_request(*handles)
        else:
            raise RuntimeError('Unknown task: {!r} from {!r}'.format(task, current))

        return loop()

    def after_loop():
        if PPINFO:
            logdebug("state %s", ppinfo())

        # pulling drivers, i.e. those explicitly run, and drivers for
        # persistent nodes must finish.
        unfinished = [(driver.key_abbrev, waitfor) for waitfor, lst in waiting.items()
                      for driver in lst
                      if driver in pulling or driver.node in persistent]

        if unfinished:
            logdebug("state %s", ppinfo())
            logerror('The following nodes did not finish:\n  %s',
                     '\n  '.join('{} waiting for {}'.format(*x) for x in
                                 sorted(unfinished)))
            if RAISE_IF_UNFINISHED:
                assert not unfinished, sorted(unfinished)

    return streams, loop, process_task, after_loop



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import inspect
import os
import random
import shutil
import tempfile
import unittest
from itertools import count
from logging import getLogger

import marv.app
import marv.model
from marv.model import db
from marv.site import Site


KEEP = os.environ.get('KEEP')
log = getLogger(__name__)


def scan_foo(directory, subdirs, filenames):
    return [(os.path.basename(x), [x]) for x in filenames
            if x.endswith('.foo')]


def scan_bar(directory, subdirs, filenames):
    return [(os.path.basename(x), [x]) for x in filenames
            if x.endswith('.bar')]


class TestCase(unittest.TestCase):
    CONFIG = """
    [marv]
    collections = foo bar

    [collection foo]
    scanner = marv.tests.test_site:scan_foo
    scanroots = scanroot/foo

    [collection bar]
    scanner = marv.tests.test_site:scan_bar
    scanroots = scanroot/bar
    """
    counter = count()

    def setUp(self):
        self.sitedir = sitedir = tempfile.mkdtemp()
        self.scanroot = scanroot = os.path.join(sitedir, 'scanroot')
        os.mkdir(scanroot)
        os.mkdir(os.path.join(scanroot, 'foo'))
        os.mkdir(os.path.join(scanroot, 'bar'))
        self.siteconf = siteconf = os.path.join(sitedir, 'marv.conf')
        with open(siteconf, 'w') as f:
            f.write(inspect.cleandoc(self.CONFIG))

        prefix = 'test{}_'.format(self.counter.next())
        marv.model._LISTING_PREFIX
        marv.model._LISTING_PREFIX = prefix
        self.site = Site(siteconf)

        app = marv.app.create_app(self.site)
        appctx = app.app_context()
        appctx.push()

        def cleanup():
            appctx.pop()
            db.session.remove()
            del self.site
            for table in [v for k, v in db.metadata.tables.items()
                          if k.startswith(prefix)]:
                db.metadata.remove(table)
            if not KEEP:
                shutil.rmtree(sitedir)
            else:
                print('Keeping {}'.format(sitedir))
        self.cleanup = cleanup

    def tearDown(self):
        self.cleanup()

    def generate_foo(self, name):
        filename = os.path.join(self.scanroot, 'foo', '{}.foo'.format(name))
        with open(filename, 'w') as f:
            f.write('')
        log.verbose('wrote %s', filename)
        return filename

    def generate_bar(self, name):
        filename = os.path.join(self.scanroot, 'bar', '{}.bar'.format(name))
        with open(filename, 'w') as f:
            f.write('')
        log.verbose('wrote %s', filename)
        return filename

    def test_flow_query_and_tag(self):
        # init and scan empty
        site = self.site
        site.init()
        site.scan()
        self.assertEqual(site.query(), [])

        # generate some
        random.seed(42)
        foo1 = self.generate_foo('foo1')
        foo2 = self.generate_foo('foo2')
        bar1 = self.generate_bar('bar1')
        site.scan()
        self.assertEqual(site.query(), ['k5jgqruqhoyt5xsweq46tqnyem',
                                        'tv43di37ggabzui2m4dpwqgwxu',
                                        'vfqitpfhd46ru3jnhsw3gzu4xu'])

        # generate more and rescan
        foo3 = self.generate_foo('foo3')
        bar2 = self.generate_bar('bar2')
        site.scan()
        self.assertEqual(len(site.query()), 5)

        # get ids
        foo1id = site.query(path=foo1)[0]
        foo2id = site.query(path=foo2)[0]
        foo3id = site.query(path=foo3)[0]
        bar1id = site.query(path=bar1)[0]
        bar2id = site.query(path=bar2)[0]

        # query combinations
        self.assertEqual(set(site.query(collections=['bar'])), {bar1id, bar2id})
        self.assertEqual(site.query(path=foo1[:-1]), [foo1id])
        self.assertEqual(site.query(collections=['bar'], path=foo1[:-1]), [])
        self.assertEqual(site.query(collections=['foo', 'bar'], path=foo1), [foo1id])

        # tag
        self.site.tag([foo1id, foo2id, bar1id], add=['a', 'b'])
        self.site.tag([foo3id], add=['c'])
        self.assertEqual(site.listtags(), ['a', 'b', 'c'])
        self.assertEqual(site.listtags(collections=['foo']), ['a', 'b', 'c'])
        self.assertEqual(site.listtags(collections=['bar']), ['a', 'b'])

        # query tagged
        self.assertEqual(set(site.query(tags=['a'])), {foo1id, foo2id, bar1id})
        self.assertEqual(set(site.query(tags=['b', 'c'])),
                         {foo1id, foo2id, foo3id, bar1id})

        # untag
        self.site.tag([foo1id, bar1id, bar2id], add=['x'], remove=['a', 'b'])
        self.assertEqual(site.listtags(), ['a', 'b', 'c', 'x'])
        self.assertEqual(site.listtags(collections=['foo']), ['a', 'b', 'c', 'x'])
        self.assertEqual(site.listtags(collections=['bar']), ['a', 'b', 'x'])
        self.assertEqual(site.query(path=foo1, tags=['a']), [])
        self.assertEqual(site.query(path=foo2, tags=['a']), [foo2id])
        self.site.tag([foo1id], remove=['x'])

        # cleanup tags
        self.site.cleanup_tags()
        self.assertEqual(site.listtags(), ['a', 'b', 'c', 'x'])
        self.assertEqual(site.listtags(collections=['foo']), ['a', 'b', 'c'])
        self.assertEqual(site.listtags(collections=['bar']), ['x'])

        # run nodes
        self.site.run(foo1id)

    def test_outdated_node(self):
        site = self.site

        site.init()
        foo1 = self.generate_foo('foo1')
        site.scan()
        foo1id = site.query(path=foo1)[0]



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import functools
import json
import os
import re
import sys
import time
from collections import Mapping, OrderedDict, defaultdict, namedtuple
from functools import partial
from inspect import getmembers
from itertools import cycle, groupby
from logging import getLogger

from sqlalchemy.sql import column, func, select

from marv import utils
from marv.config import ConfigError, calltree, getdeps, make_funcs, parse_function
from marv.model import Comment, Dataset, File, STATUS, Tag, dataset_tag, db
from marv.model import make_listing_model
from marv_detail import FORMATTER_MAP, detail_to_dict
from marv_detail.types_capnp import Detail
from marv_node.setid import SetID
from marv_store import Store


FILTER_OPERATORS = \
    'lt le eq ne ge gt substring startswith any all substring_any words'.split()
FILTER_MAP = {
    'datetime': lambda ns: None if ns is None else int(ns / 10**6),
    'filesize': lambda x: None if x is None else int(x),
    'float': lambda x: None if x is None else float(x),
    'int': lambda x: None if x is None else int(x),
    'subset': lambda lst: [unicode(x) for x in lst or []],
    'string': lambda x: None if x is None else unicode(x),
    'string[]': lambda lst: [unicode(x) for x in lst or []],
    'timedelta': lambda ns: None if ns is None else int(ns / 10**6),
    'words': lambda lst: ' '.join(lst or []),
}

Filter = namedtuple('Filter', 'name value operator type')
FilterSpec = namedtuple('FilterSpec', 'name title operators value_type function')
ListingColumn = namedtuple('ListingColumn', 'name heading formatter islist function')
SummaryItem = namedtuple('SummaryItem', 'id title formatter islist function')


def make_filter_spec(line):
    fields = [x.strip() for x in line.split('|', 4)]
    name, title, ops, value_type, function = fields
    ops = ops.split()
    return FilterSpec(name, title, ops, value_type, function)


def make_listing_column(line):
    fields = [x.strip() for x in line.split('|', 3)]
    name, heading, formatter, function = fields
    islist = formatter.endswith('[]')
    if islist:
        formatter = formatter[:-2]
    return ListingColumn(name, heading, formatter, islist, function)


def make_summary_item(line):
    fields = [x.strip() for x in line.split('|', 3)]
    id, title, formatter, function = fields
    islist = formatter.endswith('[]')
    if islist:
        formatter = formatter[:-2]
    return SummaryItem(id, title, formatter, islist, function)


def flatten_syntax(functree):
    functree = (flatten_syntax(x) if isinstance(x, tuple) else x for x in functree)
    functree = (x if len(x) > 1 else x[0] for x in functree)
    functree = ({} if x[1] == [] else
                ('map', {}, ('idx', x[1][0]), x[1][1]) if x[0] == 'rows' else
                x for x in functree)
    return tuple(functree)


def parse_summary(items):
    summary = []
    for item in items:
        functree, pos = parse_function(item.function)
        assert pos == len(item.function)
        dct = item._asdict()
        functree = flatten_syntax(functree)
        dct['function'] = functree
        summary.append(dct)
    return summary


class UnknownOperator(Exception):
    pass


def esc(s):
    return s.replace('$', '$$')\
            .replace('_', '$_')\
            .replace('%', '$%')


rowdumps = partial(json.dumps, sort_keys=True, separators=(',', ':'))


class Collections(Mapping):
    _collections = None

    @property
    def default_id(self):
        return self._dct.iterkeys().next()

    @property
    def _dct(self):
        if self._collections:
            return self._collections
        self._collections = self.loadconfig(self.config)
        return self._collections

    @staticmethod
    def loadconfig(config):
        names = config.marv.collections
        collections = OrderedDict((x, Collection(config, x)) for x in names)
        assert names == collections.keys(), (names, collections.keys())
        scanroots = [x for collection in collections.values()
                     for x in collection.scanroots]
        assert len(set(scanroots)) == len(scanroots),\
            "Scanroots must not be shared between collections"
        return collections

    def __init__(self, config):
        self.config = config

    def __dir__(self):
        return list(self._dct.viewkeys() | self.__dict__.viewkeys() |
                    {x[0] for x in getmembers(type(self))})

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, key):
        return self._dct[key]

    def __iter__(self):
        return iter(self._dct)

    def __len__(self):
        return len(self._dct)


def cached_property(func):
    """Create read-only property that caches its function's value"""
    @functools.wraps(func)
    def cached_func(self):
        cacheattr = '_{}'.format(func.func_name)
        try:
            return getattr(self, cacheattr)
        except AttributeError:
            value = func(self)
            setattr(self, cacheattr, value)
            return value
    return property(cached_func)


class Collection(object):
    @property
    def compare(self):
        # TODO: should we load?
        return self.section.compare

    @cached_property
    def detail_deps(self):
        deps = set()
        deps.update(getdeps(self.detail_title))
        deps.update(self.section.detail_sections)
        deps.update(self.section.detail_summary_widgets)
        deps.difference_update(['comments', 'dataset', 'status', 'tags'])
        return deps

    @cached_property
    def detail_sections(self):
        nodes = self.nodes
        return [nodes[x] for x in self.section.detail_sections]

    @cached_property
    def detail_summary_widgets(self):
        nodes = self.nodes
        return [nodes[x] for x in self.section.detail_summary_widgets]

    @cached_property
    def detail_title(self):
        return self.section.detail_title

    @cached_property
    def filter_functions(self):
        funcs = []
        for spec in self.filter_specs.values():
            if spec.name in ('comments', 'status', 'tags'):
                continue
            functree, pos = parse_function(spec.function)
            assert pos == len(spec.function)
            funcs.append((spec, functree))
        return funcs

    @cached_property
    def filter_specs(self):
        specs = [make_filter_spec(line) for line in self.section.filters]
        names = [x.name for x in specs]
        assert all(re.match('^[a-z0-9_]+$', x) for x in names)
        assert 'setid' in names
        assert 'tags' in names
        return OrderedDict((x.name, x) for x in specs)

    @cached_property
    def listing_columns(self):
        return [make_listing_column(line) for line in self.section.listing_columns]

    @cached_property
    def listing_deps(self):
        deps = {x for y in self.filter_functions for x in getdeps(y[1])}
        deps.update(x for y in self.listing_functions for x in getdeps(y[1]))
        deps.difference_update(['comments', 'dataset', 'status', 'tags'])
        return deps

    @cached_property
    def listing_functions(self):
        funcs = []
        for col in self.listing_columns:
            functree, pos = parse_function(col.function)
            assert pos == len(col.function)
            funcs.append((col, functree))
        return funcs

    @cached_property
    def model(self):
        return make_listing_model(self.name, self.filter_specs)

    @cached_property
    def nodes(self):
        nodes = OrderedDict()
        linemap = {}
        for line in self.section.nodes:
            try:
                nodename, node = utils.find_obj(line, True)
            except AttributeError:
                raise ConfigError(self.section, 'nodes', 'Cannot find node %s' % line)
            if node in linemap:
                raise ConfigError(self.section, 'nodes',
                                  '%s already listed as %s' %
                                  (line, linemap[node]))
            linemap[node] = line
            if nodename in nodes:
                raise ConfigError(self.section, 'nodes', 'duplicate name %s' % (nodename,))
            if not node.schema:
                raise ConfigError(self.section, 'nodes', '%s does not define a schema!' % (line,))
            nodes[nodename] = node
        return nodes

    @property
    def scanner(self):
        return utils.find_obj(self.section.scanner)

    @property
    def scanroots(self):
        return self.section.scanroots

    @cached_property
    def sortcolumn(self):
        listing_sort = self.section.listing_sort
        try:
            return 0 if not listing_sort[0] else \
                (i for i, x in enumerate(self.listing_columns)
                 if x.name == listing_sort[0]).next()
        except StopIteration:
            raise ValueError("No column named '%s'" % listing_sort[0])

    @cached_property
    def sortorder(self):
        listing_sort = self.section.listing_sort
        try:
            sortorder = listing_sort[1]
            orders = ['ascending', 'descending']
            return {x: x for x in orders}[sortorder]
        except IndexError:
            return 'ascending'
        except KeyError:
            raise ConfigError(self.section, 'listing_sort',
                              '{} not in [{}]'.format(sortorder, ', '.join(orders)))

    @cached_property
    def summary_items(self):
        items = [make_summary_item(line)
                 for line in self.section.listing_summary]
        return parse_summary(items)

    @property
    def section(self):
        return self.config['collection {}'.format(self.name)]

    def __init__(self, config, name):
        self.config = config
        self.name = name

    def filtered_listing(self, filters):
        model = self.model
        Listing = model.Listing
        listing = Listing.__table__
        relations = model.relations
        secondaries = model.secondaries
        stmt = select([Listing.id.label('id'),
                       Listing.row.label('row'),
                       Dataset.status.label('status'),
                       Tag.value.label('tag_value')])\
                       .select_from(listing.outerjoin(Dataset)
                                    .outerjoin(dataset_tag)
                                    .outerjoin(Tag))\
                       .where(Dataset.discarded.isnot(True))
        for name, value, op, val_type in filters:
            if isinstance(value, long):
                value = min([value, sys.maxsize])

            if name == 'comments':
                containstext = Comment.text.like('%{}%'.format(esc(value)), escape='$')
                commentquery = db.session.query(Comment.dataset_id)\
                                         .filter(containstext)\
                                         .group_by(Comment.dataset_id)
                stmt = stmt.where(Listing.id.in_(commentquery.subquery()))
                continue
            elif name == 'status':
                status_ids = STATUS.keys()
                bitmasks = [2**status_ids.index(x) for x in value]
                if op == 'any':
                    stmt = stmt.where(reduce(lambda x, y: x|y,
                                             (Dataset.status.op('&')(x) for x in bitmasks)))
                elif op == 'all':
                    bitmask = sum(bitmasks)
                    stmt = stmt.where(Dataset.status.op('&')(bitmask) == bitmask)
                else:
                    raise UnknownOperator(op)
                continue
            elif name == 'tags':
                if op == 'any':
                    relquery = db.session.query(dataset_tag.c.dataset_id)\
                                         .join(Tag)\
                                         .filter(Tag.value.in_(value))
                    stmt = stmt.where(Listing.id.in_(relquery.subquery()))
                elif op == 'all':
                    relquery = Tag.query\
                                  .join(dataset_tag)\
                                  .filter(reduce(lambda x, y: x|y, (Tag.value == x for x in value)))\
                                  .group_by(dataset_tag.c.dataset_id)\
                                  .having(db.func.count('*') == len(value))\
                                  .with_entities(dataset_tag.c.dataset_id)
                    stmt = stmt.where(Listing.id.in_(relquery.subquery()))
                else:
                    raise UnknownOperator(op)
                continue
            elif val_type == 'datetime':
                if op == 'eq':
                    col = getattr(Listing, name)
                    stmt = stmt.where(col.between(value, value + 24 * 3600 * 1000))
                    continue
                elif op == 'ne':
                    col = getattr(Listing, name)
                    stmt = stmt.where(~col.between(value, value + 24 * 3600 * 1000))
                    continue
                elif op in ['le', 'gt']:
                    value = value + 24 * 3600 * 1000

            col = getattr(Listing, name)
            if op == 'lt':
                stmt = stmt.where(col < value)
            elif op == 'le':
                stmt = stmt.where(col <= value)
            elif op == 'eq':
                stmt = stmt.where(col == value)
            elif op == 'ne':
                stmt = stmt.where(col != value)
            elif op == 'ge':
                stmt = stmt.where(col >= value)
            elif op == 'gt':
                stmt = stmt.where(col > value)
            elif op == 'substring':
                stmt = stmt.where(col.like('%{}%'.format(esc(value)), escape='$'))
            elif op == 'startswith':
                stmt = stmt.where(col.like('{}%'.format(esc(value)), escape='$'))
            elif op == 'any':
                rel = relations[name]
                sec = secondaries[name]
                relquery = rel.query\
                              .join(sec)\
                              .filter(rel.value.in_(value))\
                              .with_entities(sec.c.listing_id)
                stmt = stmt.where(Listing.id.in_(relquery.subquery()))
            elif op == 'all':
                rel = relations[name]
                sec = secondaries[name]
                relquery = rel.query\
                              .join(sec)\
                              .filter(reduce(lambda x, y: x|y, (rel.value == x for x in value)))\
                              .group_by(sec.c.listing_id)\
                              .having(db.func.count('*') == len(value))\
                              .with_entities(sec.c.listing_id)
                stmt = stmt.where(Listing.id.in_(relquery.subquery()))
            elif op == 'substring_any':
                rel = relations[name]
                stmt = stmt.where(col.any(rel.value.like('%{}%'.format(esc(value)),
                                                            escape='$')))
            elif op == 'words':
                stmt = reduce(lambda stmt, x: stmt.where(col.like('%{}%'.format(esc(x)),
                                                             escape='$')),
                               value, stmt)
            else:
                raise UnknownOperator(op)
        stmt = select([column('row'), column('status'), func.json_group_array(column('tag_value'))])\
               .select_from(stmt.order_by(column('tag_value')))\
               .group_by('id')
        return stmt

    def scan(self, scanpath, dry_run=False):
        Listing = self.model.Listing
        log = getLogger('.'.join([__name__, self.name]))
        scanroot = (x for x in self.scanroots if scanpath.startswith(x)).next()
        if not os.path.isdir(scanpath):
            log.warn('%s does not exist or is not a directory', scanpath)

        log.verbose("scanning %s'%s'", 'dry_run ' if dry_run else '', scanpath)

        # missing/changed flag for known files
        startswith = File.path.like('{}%'.format(esc(scanpath)), escape='$')
        known_files = File.query.filter(startswith)\
                                .join(Dataset)\
                                .filter(Dataset.discarded.isnot(True))
        known_filenames = defaultdict(set)
        changes = defaultdict(list)  # all mtime/missing changes in one transaction
        for file in known_files:
            path = file.path
            known_filenames[os.path.dirname(path)].add(os.path.basename(path))
            try:
                mtime = utils.mtime(path)
                missing = False
            except OSError:
                mtime = None
                missing = True
            if missing ^ bool(file.missing):
                log.info("%s '%s'", 'lost' if missing else 'recovered', path)
                changes[file.dataset_id].append((file, missing))
            if mtime and int(mtime * 1000) > file.mtime:
                log.info("mtime newer '%s'", path)
                changes[file.dataset_id].append((file, mtime))

        # Apply missing/mtime changes
        if not dry_run and changes:
            ids = changes.keys()
            for dataset in Dataset.query.filter(Dataset.id.in_(ids)):
                for file, change in changes.pop(dataset.id):
                    check_outdated = False
                    if type(change) is bool:
                        file.missing = change
                        dataset.missing = change
                    else:
                        file.mtime = int(change * 1000)
                        check_outdated = True
                if check_outdated:
                    self._check_outdated(dataset)
                dataset.time_updated = int(utils.now())
            assert not changes
            db.session.commit()

        # Scan for new files
        batch = []
        for directory, subdirs, filenames in os.walk(scanpath):
            # Ignore directories containing a .marvignore file
            if os.path.exists(os.path.join(directory, '.marvignore')):
                subdirs[:] = []
                continue

            # Ignore hidden directories and traverse subdirs alphabetically
            subdirs[:] = sorted([x for x in subdirs if x[0] != '.'])

            # Ignore hidden and known files
            known = known_filenames[directory]
            filenames = {x for x in filenames if x[0] != '.'}
            filenames = sorted(filenames - known)

            for name, files in self.scanner(directory, subdirs, filenames):
                files = [x if os.path.isabs(x) else os.path.join(directory, x)
                         for x in files]
                assert all(x.startswith(directory) for x in files), files
                if dry_run:
                    log.info("would add '%s': '%s'", directory, name)
                else:
                    dataset = self.make_dataset(files, name)
                    batch.append(dataset)
                    if len(batch) > 50:
                        self._add_batch(log, batch)

        if not dry_run and batch:
            self._add_batch(log, batch)

        log.verbose("finished %s'%s'", 'dry_run ' if dry_run else '', scanpath)

    def _check_outdated(self, dataset):
        storedir = self.config.marv.storedir
        setdir = os.path.join(storedir, str(dataset.setid))
        latest = [os.path.realpath(x)
                  for x in [os.path.join(setdir, x) for x in os.listdir(setdir)]
                  if os.path.islink(x)]
        oldest_mtime = utils.mtime(os.path.join(setdir, 'detail.json'))
        for nodedir in latest:
            for dirpath, dirnames, filenames in os.walk(nodedir):
                for name in filenames:
                    path = os.path.join(dirpath, name)
                    oldest_mtime = min(oldest_mtime, utils.mtime(path))
        dataset_mtime = max(x.mtime for x in dataset.files)
        dataset.outdated = int(oldest_mtime * 1000) < dataset_mtime

    def restore_datasets(self, data):
        log = getLogger('.'.join([__name__, self.name]))
        batch = []
        comments = []
        tags = []
        for dataset in data:
            _comments = dataset.pop('comments')
            _tags = dataset.pop('tags')
            dataset = self.make_dataset(_restore=True, **dataset)
            comments.extend(Comment(dataset=dataset, **x) for x in _comments)
            tags.append((dataset, _tags))
            batch.append(dataset)
            if len(batch) > 50:
                db.session.add_all(comments)
                comments[:] = []
                self._add_batch(log, batch)
                self._add_tags(tags)
        db.session.add_all(comments)
        comments[:] = []
        self._add_batch(log, batch)
        self._add_tags(tags)

    def _add_tags(self, data):
        tags = {tag for _, tags in data for tag in tags}
        stmt = Tag.__table__.insert().prefix_with('OR IGNORE')
        db.session.execute(stmt, [{'collection': self.name,
                                   'value': x} for x in tags])

        mapping = {value: id for id, value in (db.session.query(Tag.id, Tag.value)
                                               .filter(Tag.collection == self.name)
                                               .filter(Tag.value.in_(tags)))}

        stmt = dataset_tag.insert()
        values = [{'tag_id': mapping[x], 'dataset_id': y}
                  for tags, id in [(tags, dataset.id) for dataset, tags in data]
                  for x, y in zip(tags, cycle([id]))]
        db.session.execute(stmt, values)
        db.session.commit()
        data[:] = []

    def _add_batch(self, log, batch):
        Listing = self.model.Listing
        relations = self.model.relations

        # We need ids to render the listings; flush would result in
        # longer write transaction.
        db.session.add_all(batch)
        db.session.commit()
        for dataset in batch:
            log.info('added %r', dataset)

        queue = []
        relvalues = defaultdict(set)
        listings = []
        for dataset in batch:
            row, fields, relfields = self.render_listing(dataset)
            listing = Listing(dataset=dataset, row=rowdumps(row), **fields)
            listings.append(listing)
            for key, values in relfields.items():
                if not values:
                    continue
                relvalues[key].update(values or [])
                queue.append((listing, key, values))

        relmap = {}
        for key, values in relvalues.items():
            if not values:
                relmap[key] = {}
                continue
            Rel = relations[key]
            insert = Rel.__table__.insert().prefix_with('OR IGNORE')
            db.session.execute(insert, [{'value': x} for x in values])
            query = db.session.query(Rel)\
                              .filter(Rel.value.in_(values))
            relmap[key] = {rel.value: rel for rel in query}

        for listing, key, values in queue:
            rels = relmap[key]
            setattr(listing, key, [rels[x] for x in values])
        db.session.add_all(listings)
        db.session.commit()
        batch[:] = []

    def make_dataset(self, files, name, time_added=None, discarded=None, setid=None, status=None,
                     timestamp=None, _restore=None):
        setid = setid or SetID.random()
        if _restore:
            files = [File(idx=i, **x) for i, x in enumerate(files)]
        else:
            files = [File(idx=i, mtime=int(utils.mtime(path) * 1000), path=path, size=stat.st_size)
                     for i, (path, stat)
                     in enumerate((path, os.stat(path)) for path in files)]
        time_added = int(utils.now() * 1000) if time_added is None else time_added
        dataset = Dataset(collection=self.name,
                          files=files,
                          name=name,
                          discarded=discarded,
                          status=status,
                          time_added=time_added,
                          timestamp=timestamp or max(x.mtime for x in files),
                          setid=setid)

        storedir = self.config.marv.storedir
        store = Store(storedir, self.nodes)
        store.add_dataset(dataset, exists_okay=_restore)
        self.render_detail(dataset)
        return dataset

    def render_detail(self, dataset):
        storedir = self.config.marv.storedir
        setdir = os.path.join(storedir, str(dataset.setid))
        try:
            os.mkdir(setdir)
        except OSError:
            pass
        assert os.path.isdir(setdir), setdir
        store = Store(storedir, self.nodes)
        funcs = make_funcs(dataset, setdir, store)

        summary_widgets = [
            x[0]._reader for x in
            [store.load(setdir, node, default=None) for node in self.detail_summary_widgets]
            if x
        ]

        sections = [
            x[0]._reader for x in
            [store.load(setdir, node, default=None) for node in self.detail_sections]
            if x
        ]

        kw = {'title': calltree(self.detail_title, funcs),
              'sections': sections,
              'summary': {'widgets': summary_widgets}}
        detail = Detail.new_message(**kw).as_reader()
        dct = detail_to_dict(detail)
        fd = os.open(os.path.join(setdir, '.detail.json'),
                     os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666)
        jsonfile = os.fdopen(fd, 'w')
        json.dump(dct, jsonfile, sort_keys=True)
        jsonfile.close()
        os.rename(os.path.join(setdir, '.detail.json'),
                  os.path.join(setdir, 'detail.json'))
        self._check_outdated(dataset)

    def render_listing(self, dataset):
        storedir = self.config.marv.storedir
        setdir = os.path.join(storedir, str(dataset.setid))
        store = Store(storedir, self.nodes)
        funcs = make_funcs(dataset, setdir, store)

        values = []
        for col, functree in self.listing_functions:
            value = calltree(functree, funcs)
            if value is not None:
                transform = FORMATTER_MAP[col.formatter + ('[]' if col.islist else '')]
                value = transform(value)
            values.append(value)
        row = {'id': dataset.id,
               'setid': str(dataset.setid),
               'tags': ['#TAGS#'],
               'values': values}

        fields = {}
        relfields = {}
        relations = self.model.relations
        for filter_spec, functree in self.filter_functions:
            value = calltree(functree, funcs)
            transform = FILTER_MAP[filter_spec.value_type]
            value = transform(value)
            target = relfields if filter_spec.name in relations else fields
            target[filter_spec.name] = value

        return row, fields, relfields

    def update_listings(self, datasets):
        assert datasets
        # TODO: similar to _add_batch
        relations = self.model.relations
        secondaries = self.model.secondaries

        queue = []
        relvalues = defaultdict(set)
        render_listing = self.render_listing
        rendered = [render_listing(x) for x in datasets]
        listings = []
        for i, (row, fields, relfields) in enumerate(rendered):
            assert 'id' not in fields
            assert 'row' not in fields
            dataset = datasets[i]
            fields['id'] = listing_id = dataset.id
            fields['row'] = rowdumps(row)
            listings.append(fields)
            for key, values in relfields.items():
                relvalues[key].update(values or [])
                queue.append((listing_id, key, values))

        # Update new relation values in bulk per relation and generate
        # relation map from value to id
        relmap = {}
        for key, values in relvalues.items():
            if not values:
                relmap[key] = {}
                continue
            relation = relations[key].__table__
            stmt = relation.insert()\
                           .prefix_with('OR IGNORE')
            db.session.execute(stmt, [{'value': x} for x in values])
            query = db.session.query(relation.c.value, relation.c.id)\
                              .filter(relation.c.value.in_(values))
            relmap[key] = {value: id for value, id in query}

        # bulk delete associations per relation
        for key, listing_ids in [(key, [x[0] for x in group])
                         for key, group in groupby(queue, lambda x: x[1])]:
            secondary = secondaries[key]
            stmt = secondary.delete()\
                            .where(secondary.c.listing_id.in_(listing_ids))
            db.session.execute(stmt)

        # bulk insert/replace listings
        stmt = self.model.Listing.__table__.insert()\
                                           .prefix_with('OR REPLACE')
        db.session.execute(stmt, listings)

        # bulk insert associations per relation
        for listing_id, key, values in queue:
            if not values:
                continue
            assert isinstance(values, list), (listing_id, key, values)
            relids = relmap[key]
            secondary = secondaries[key]
            stmt = secondary.insert()\
                            .prefix_with('OR IGNORE')
            db.session.execute(stmt, [{'listing_id': listing_id,
                                       'relation_id': relids[x]} for x in values])

        db.session.commit()



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import json
import sys

import capnp

from marv_pycapnp import Wrapper
from .types_capnp import Detail, Widget


FORMATTER_MAP = {
    'date': lambda ns: None if ns is None else int(ns / 10**6),
    'datetime': lambda ns: None if ns is None else int(ns / 10**6),
    'filesize': lambda x: None if x is None else int(x),
    'float': float,
    'icon[]': lambda x: x,
    'int': int,
    'link': lambda x: x,
    'pill[]': lambda x: x,
    'rellink': lambda x: x,
    'route': lambda x: x,
    'string': lambda x: x,
    'timedelta': lambda ns: None if ns is None else int(ns / 10**6),
}


def detail_to_dict(obj):
    dct = obj.to_dict(verbose=True, which=True)
    widgets = dct.get('summary', {'widgets': []})['widgets'][:]
    widgets.extend(widget for sec in dct['sections'] for widget in sec['widgets'])
    for widget in widgets:
        try:
            fixup_widget(widget)
        except:
            from pprint import pformat
            print(pformat(dct), file=sys.stderr)
            print(pformat(widget), file=sys.stderr)
            raise
    return dct


def make_map_dict(dct):
    mapdct = Wrapper.from_dict(Widget.Map, dct)._reader.to_dict(verbose=True, which=True)
    fixup_map(mapdct)
    return mapdct


def fixup_widget(dct):
    which = dct['type'] = dct.pop('_which')
    data = dct['data'] = dct.pop(which)
    if which == 'custom':
        dct['type'] = data['type']
        dct['data'] = json.loads(data['data'])
    elif which == 'keyval':
        for item in data['items']:
            cell = item.pop('cell')
            transform = FORMATTER_MAP[item['formatter'] + ('[]' if item['list'] else '')]
            item['value'] = transform(cell[cell['_which']])
    elif which == 'link':
        if not data['download']:
            del data['download']
    elif which == 'map':
        fixup_map(data)
    elif which == 'mapPartial':
        dct['type'] = 'map'
        assert data.startswith('marv-partial:'), data
    elif which == 'mpld3' and not data.startswith('marv-partial:'):
        dct['data'] = json.loads(data)
    elif which == 'pointcloud':
        transform = data['transform']
        assert len(transform) == 16, transform
    elif which == 'table':
        formatter = [FORMATTER_MAP[col['formatter'] + ('[]' if col['list'] else '')]
                     for col in data['columns']]
        for row in data['rows']:
            row['values'] = values = []
            for idx, cell in enumerate(row['cells']):
                values.append(formatter[idx](cell[cell['_which']]))
            del row['cells']
        for action in data['actions']:
            action['data'] = json.loads(action['data'])


def fixup_map(data):
    zoom = data['zoom']
    data['zoom'] = [zoom['min'], zoom['max']]
    for layer in data['layers']:
        transform = layer['transform']
        assert len(transform) == 16, transform
        layer['type'] = layertype = layer.pop('_which')
        if layertype == 'tiles':
            for tile in layer['tiles']:
                zoom = tile['zoom']
                tile['zoom'] = [zoom['min'], zoom['max']]
        elif layertype == 'geojson':
            fixup_geojson(layer['geojson'])
        else:
            raise RuntimeError(layertype)


def fixup_geojson(geojson):
    geotype = geojson.pop('_which')
    geojson['type'] = geotype[0].upper() + geotype[1:]
    geojson.update(geojson.pop(geotype))
    if geotype == 'feature':
        fixup_geometry(geojson['geometry'])
    elif geotype == 'featureCollection':
        for feature in geojson['features']:
            feature['type'] = 'Feature'
            fixup_geometry(feature['geometry'])
            properties = feature['properties']
            timestamps = properties['timestamps']
            if timestamps:
                properties['timestamps'] = [int(x / 1e6) for x in timestamps]
            else:
                del properties['timestamps']
            if not properties['rotations']:
                del properties['rotations']
    elif geotype == 'geometryCollection':
        for geo in geojson['geometries']:
            fixup_geometry(geo)
    elif geotype == 'polygon':
        fixup_polygon(geojson['coordinates'])
    elif geotype == 'multiPolygon':
        for poly in geojson['coordinates']:
            fixup_polygon(poly)


def fixup_polygon(linear_rings):
    first = linear_rings[0]
    if not ccw(first):
        first.reverse()
    for coords in linear_rings[1:]:
        if ccw(coords):
            coords.reverse()


def fixup_geometry(geometry):
    which = geometry.pop('_which')
    geometry['type'] = which[0].upper() + which[1:]
    geometry.update(geometry.pop(which))
    if which == 'polygon':
        fixup_polygon(geometry['coordinates'])
    elif which == 'multiPolygon':
        for poly in geometry['coordinates']:
            fixup_polygon(poly)


def ccw(points):
    return sum((x2 + x1) / 2 * (y2 - y1) for (x1, y1), (x2, y2) in
               (points[i:i+2] for i in range(len(points) - 1))) > 0



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import json
import mimetypes
import os

import flask
from flask import current_app

from marv.model import Comment, Dataset, File, Tag, dataset_tag, db
from marv_node.setid import SetID
from .tooling import api_group as marv_api_group


@marv_api_group()
def dataset(app):
    pass


@dataset.endpoint('/file-list', methods=['POST'])
def file_list():
    ids = flask.request.get_json()
    if not ids:
        flask.abort(400)

    # TODO: remove scanroot prefix?

    query = db.session.query(File.path)\
                      .filter(File.dataset_id.in_(ids))\
                      .order_by(File.path)
    paths = [x[0] for x in query]

    query = db.session.query(Dataset.setid, File.idx)\
                      .filter(Dataset.id.in_(ids))\
                      .join(File)\
                      .order_by(Dataset.setid)
    urls = ['dataset/{}/{}'.format(setid, idx)
            for setid, idx in query]

    # TODO: complain about invalid/unknown ids?

    return flask.jsonify({'paths': paths, 'urls': urls})


def _send_detail_json(setid, setdir):
    try:
        with open(os.path.join(setdir, 'detail.json')) as f:
            detail = json.load(f)
    except IOError:
        return flask.abort(404)

    # TODO: investigate merge into one
    dataset_id, collection = db.session.query(Dataset.id, Dataset.collection)\
                                       .filter(Dataset.setid == setid)\
                                       .one()
    comments = db.session.query(Comment.author, Comment.time_added, Comment.text)\
                         .filter(Comment.dataset_id == dataset_id)\
                         .order_by(Comment.time_added)
    detail['comments'] = [{'author': x[0],
                           'timeAdded': x[1],
                           'text': x[2]} for x in comments]

    collection = current_app.site.collections[collection]
    alltags = db.session.query(Tag.value)\
                        .filter(Tag.collection == collection.name)\
                        .order_by(Tag.value)
    detail['all_known_tags'] = [x[0] for x in alltags]

    tags = db.session.query(Tag.value)\
                     .join(dataset_tag)\
                     .filter(dataset_tag.c.dataset_id == dataset_id)\
                     .order_by(Tag.value)
    detail['tags'] = [x[0] for x in tags]
    detail['collection'] = collection.name
    detail['id'] = dataset_id
    detail['setid'] = setid
    resp = flask.jsonify(detail)
    resp.headers['Cache-Control'] = 'no-cache'
    return resp


@dataset.endpoint('/dataset/<setid>', defaults={'path': 'detail.json'})
@dataset.endpoint('/dataset/<setid>/<path:path>')
def detail(setid, path):
    try:
        setid = str(SetID(setid))
    except TypeError:
        flask.abort(404)

    setdir = os.path.join(current_app.site.config.marv.storedir, setid)

    if path == 'detail.json':
        return _send_detail_json(setid, setdir)

    if path.isdigit():
        path = db.session.query(File.path)\
                         .join(Dataset)\
                         .filter(Dataset.setid == setid)\
                         .filter(File.idx == int(path))\
                         .scalar()
    else:
        path = flask.safe_join(setdir, path)

    # Make sure path exists and is safe
    if not os.path.isabs(path) \
       or path != os.path.normpath(path) \
       or not os.path.isfile(path):
        return flask.abort(404)

    if current_app.site.config.marv.reverse_proxy == 'nginx':
        resp = flask.make_response()
        mime = mimetypes.guess_type(path)
        resp.headers['content-type'] = \
            mime[0] if mime[0] else 'application/octet-stream'
        resp.headers['cache-control'] = 'no-cache'
        resp.headers['x-accel-buffering'] = 'no'
        resp.headers['x-accel-redirect'] = \
            (current_app.config['APPLICATION_ROOT'] or '') + path
        resp.headers.add('Content-Disposition', 'attachment',
                         filename=os.path.basename(path))
        return resp

    try:
        resp = flask.send_file(path, as_attachment=True, conditional=True)
        resp.headers['Cache-Control'] = 'no-cache'
        return resp
    except ValueError:
        flask.abort(404)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import os
import re
import time
from datetime import datetime, timedelta
from datetime import tzinfo as tzinfo_base
from importlib import import_module
from itertools import islice

from marv_node.setid import decode_setid, encode_setid


def chunked(iterable, chunk_size):
    it = iter(iterable)
    return iter(lambda: tuple(islice(it, chunk_size)), ())


def find_obj(objpath, name=False):
    modpath, objname = objpath.split(':')
    mod = import_module(modpath)
    obj = getattr(mod, objname)
    return (objname, obj) if name else obj


def mtime(path):
    """Wrapper for st_mtime for ease of mocking"""
    return os.stat(path).st_mtime


def now():
    """Wrapper for time.time for ease of mocking"""
    return time.time()


def parse_filesize(string):
    val, unit = re.match(r'^\s*([0-9.]+)\s*([kmgtpezy]b?)?\s*$', string, re.I)\
                  .groups()
    val = float(val)
    if unit:
        val *= 1 << (10 * (1 + 'kmgtpezy'.index(unit.lower()[0])))
    return int(val)


def parse_datetime(s):
    class TZInfo(tzinfo_base):
        def __init__(self, offset=None):
            self.offset = offset

        def tzname(self, dt):
            return self.offset

        def utcoffset(self, dt):
            if self.offset == 'Z':
                hours, minutes = 0, 0
            else:
                hours, minutes = self.offset[1:].split(':')
            offset = timedelta(hours=int(hours), minutes=int(minutes))
            return offset if self.offset[0] == '+' else -offset

    groups = re.match(r'^(\d\d\d\d)-(\d\d)-(\d\d)T'
                     r'(\d\d):(\d\d):(\d\d)((?:[+-]\d\d:\d\d)|Z)$', s)\
              .groups()
    tzinfo = TZInfo(groups[-1])
    return datetime(*(int(x) for x in groups[:-1]), tzinfo=tzinfo)


def parse_timedelta(d):
    match = re.match(r'^\s*(?:(\d+)\s*h)?'
                     r'\s*(?:(\d+)\s*m)?'
                     r'\s*(?:(\d+)\s*s?)?\s*$', d)
    h, m, s = match.groups() if match else (None, None, None)
    return (int(h or 0) * 3600 + int(m or 0) * 60 + int(s or  0)) * 1000;


def profile(func, sort='cumtime'):
    import functools
    import pstats
    from cProfile import Profile
    pr = Profile()
    @functools.wraps(func)
    def profiled(*args, **kw):
        pr.enable()
        rv = func(*args, **kw)
        pr.disable()
        ps = pstats.Stats(pr).sort_stats(sort)
        ps.print_stats()
        return rv
    return profiled


def underscore_to_camelCase(s):
    return ''.join(x.capitalize() for x in s.split('_'))



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from marv_node.node import StreamSpec as _StreamSpec


def select(node, name):
    """Select specific stream of a node by name.

    Args:
        node: A node producing a group of streams.
        name (str): Name of stream to select.

    Returns:
        Node outputting selected stream.
    """
    return _StreamSpec(node, name)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function


def download_link(file, dataset):
    return {
        'href': 'download/{}'.format(dataset.files.index(file)),
        'target': '_blank',
        'title': file.relpath,
    }


def file_status(file):
    state = []
    if file and file.missing:
        state.append({'icon': 'hdd',
                      'title': 'This file is missing',
                      'classes': 'text-danger'})
    return state



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from collections import namedtuple
from numbers import Integral

from .mixins import Keyed, Task, Request
from .stream import Handle


class Abort(Exception):
    pass


def create_stream(name, **header):
    """Create a stream for publishing messages.

    All keyword arguments will be used to form the header.
    """
    assert isinstance(name, basestring), name
    return CreateStream(parent=None, name=name, group=False, header=header)


def create_group(name, **header):
    assert isinstance(name, basestring), name
    return CreateStream(parent=None, name=name, group=True, header=header)


def fork(name, inputs, group):
    assert isinstance(name, basestring), name
    assert inputs
    return Fork(name, inputs, group)


def get_logger():
    return GetLogger()


def get_requested():
    return GetRequested()


def get_stream(node, name='default', setid=None):
    return GetStream(setid, node, name)


def make_file(name):
    assert isinstance(name, basestring)
    return MakeFile(None, name)


def pull(handle, enumerate=False):
    """Pulls next message for handle.

    Args:
        handle: A :class:`.stream.Handle` or GroupHandle.
        enumerate (bool): boolean to indicate whether a tuple ``(idx, msg)``
            should be returned, not unlike Python's enumerate().

    Returns:
        A :class:`Pull` task to be yielded. Marv will send the
        corresponding message as soon as it is available. For groups
        this message will be a handle to a member of the
        group. Members of groups are either streams or groups.

    Examples:
        Pulling (enumerated) message from stream::

            msg = yield marv.pull(stream)
            idx, msg = yield marv.pull(stream, enumerate=True)

        Pulling stream from group and message from stream::

            stream = yield marv.pull(group)  # a group of streams
            msg = yield marv.pull(stream)

    """
    assert isinstance(handle, Handle), handle
    return Pull(handle, enumerate)


def pull_all(*handles):
    """Pulls next message of all handles."""
    return PullAll(handles)


def pull_any(handles):
    raise NotImplementedError


def push(msg):
    return Push(msg)


def set_header(**header):
    """Set the header of a stream or group."""

    # If a node is configured to have a header, the header needs to be
    # set before yielding any messages or creating group members. Once a
    # header is set, a handle is created and dependent nodes can be
    # instantiated. For streams without headers this happens right away.

    #     @marv.node(header=True)
    #     def node():
    #         yield marv.set_header(title='Title')

    # """
    return SetHeader(header)


CreateStream = namedtuple('CreateStream', 'parent name group header')
Fork = namedtuple('Fork', 'name inputs group')
GetLogger = namedtuple('GetLogger', '')
GetRequested = namedtuple('GetRequested', '')
GetStream = namedtuple('GetStream', 'setid node name')
MakeFile = namedtuple('MakeFile', 'handle name')

Pull = namedtuple('Pull', 'handle enumerate')
PullAll = namedtuple('PullAll', 'handles')
PullAny = namedtuple('PullAny', 'handles')
PullSome = namedtuple('PullSome', 'handles')
Push = namedtuple('Push', 'output')
SetHeader = namedtuple('SetHeader', 'header')

# TODO: Rename
Request.register(Pull)
Request.register(PullAll)
Request.register(PullAny)
Request.register(PullSome)
Request.register(Push)
Request.register(SetHeader)

Request.register(CreateStream)
Request.register(Fork)
Request.register(GetLogger)
Request.register(GetRequested)
Request.register(GetStream)
Request.register(MakeFile)


class Signal(Task):
    def __repr__(self):
        return type(self).__name__.upper()

class Next(Signal):
    """Instruct to send next pending task."""
    __slots__ = ()

class Paused(Signal):
    """Indicate a generator has paused."""
    __slots__ = ()

class Resume(Signal):
    """Instruct a generator to resume."""
    __slots__ = ()

class TheEnd(Signal):
    """Indicate the end of a stream, resulting in None being sent into consumers."""
    __slots__ = ()

NEXT = Next()
PAUSED = Paused()
RESUME = Resume()
THEEND = TheEnd()


class MsgRequest(Task, Keyed):
    __slots__ = ('_handle', '_idx', '__weakref__')

    @property
    def key(self):
        return (self._handle, self._idx)

    @property
    def handle(self):
        return self._handle

    @property
    def idx(self):
        return self._idx

    def __init__(self, handle, idx, requestor):
        assert isinstance(idx, Integral), idx
        self._handle = handle
        self._idx = idx
        self._requestor = requestor

    def __iter__(self):
        return iter(self.key)

    def __repr__(self):
        return 'MsgRequest({}, {!r})'.format(self._handle, self._idx)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import os
from collections import Mapping, Sequence

from capnp.lib.capnp import _DynamicEnum
from capnp.lib.capnp import _DynamicListReader
from capnp.lib.capnp import _DynamicStructReader
from flask import current_app

from marv_node.setid import SetID


def _to_dict(value, field=None, field_type=None, which=False):
    if isinstance(value, _DynamicStructReader):
        schema = value.schema
        dct = {}
        for name in schema.non_union_fields:
            dct[name] = _to_dict(getattr(value, name), field=schema.fields[name], which=which)

        if schema.union_fields:
            _which = str(value.which)
            dct[_which] = _to_dict(getattr(value, _which), field=schema.fields[_which], which=which)
            if which:
                dct['_which'] = _which

        return dct

    elif isinstance(value, _DynamicListReader):
        element_type = (field_type or field.proto.slot.type).list.elementType
        return [
            _to_dict(element, field_type=element_type, which=which) for element in value
        ]

    elif isinstance(value, str) and (field_type or field.proto.slot.type).which.raw == 12L:  # :Text
        return value.decode('utf-8')

    elif isinstance(value, _DynamicEnum):
        return value._as_str()

    return value


def _wrap(value, streamdir, setdir, field=None, field_type=None):
    if isinstance(value, _DynamicStructReader):
        return Wrapper(value, streamdir, setdir)

    elif isinstance(value, _DynamicListReader):
        if field:
            element_type = field.proto.slot.type.list.elementType
        else:
            element_type = field_type.list.elementType
        return ListWrapper(value, element_type, streamdir, setdir)

    elif isinstance(value, str) and (field_type or field.proto.slot.type).which.raw == 12L:  # :Text
        return value.decode('utf-8')

    return value


class ListWrapper(object):
    def __init__(self, list_reader, field_type, streamdir, setdir):
        assert isinstance(list_reader, _DynamicListReader), type(list_reader)
        self._field_type = field_type
        self._reader = list_reader
        self._streamdir = streamdir
        self._setdir = setdir

    def _wrap(self, item):
        return _wrap(item, self._streamdir, self._setdir, field_type=self._field_type)

    def __getitem__(self, idx):
        if isinstance(idx, slice):  # pycapnp can't handle slices
            return [self._wrap(x) for x in list(self._reader)[idx]]
        return self._wrap(self._reader[idx])

    def __eq__(self, other):
        return self[:] == other

    def __iter__(self):
        return (self._wrap(x) for x in self._reader)

    def __len__(self):
        return len(self._reader)

    def __repr__(self):
        return '[{}]'.format(', '.join([repr(x) for x in self]))


class Wrapper(object):
    def __init__(self, struct_reader, streamdir, setdir):
        assert isinstance(struct_reader, _DynamicStructReader), type(struct_reader)
        self._reader = struct_reader
        self._streamdir = streamdir  # HACK: overloaded
        self._setdir = setdir

    # @property
    # def path(self):  # HACK: overload
    #     return os.path.realpath(os.path.join(self._streamdir, self.name))

    @property
    def relpath(self):  # HACK: overload
        path = os.path.relpath(self.path.decode('utf-8'), self._setdir)
        return path.lstrip('.')

    @classmethod
    def from_dict(cls, schema, data):
        setid = data.pop('id', None)
        if isinstance(setid, SetID):
            data['id0'], data['id1'] = setid.lohi
        dct = cls._unwrap(data)
        struct_reader = schema.new_message(**dct).as_reader()
        return cls(struct_reader, None, None)

    def to_dict(self, which=None):
        return _to_dict(self._reader, which=which)

    def load(self, node):
        from marv_store import Store
        storedir = current_app.site.config.marv.storedir
        store = Store(storedir, {node.name: node})
        return store.load(self._setdir, node, default=None)

    @classmethod
    def _unwrap(cls, data):
        unwrap = cls._unwrap
        if isinstance(data, Mapping):
            return {k: unwrap(v) for k, v in data.iteritems()}
        elif isinstance(data, Sequence) and not isinstance(data, basestring):
            return [unwrap(x) for x in data]
        elif isinstance(data, Wrapper):
            return data._reader
        else:
            return data

    def __getattr__(self, name):
        parts = name.split('_')
        field_name = parts[0] + ''.join(((x[0].upper() + x[1:]) if x else '_') for x in parts[1:])

        if field_name in self._reader.schema.fieldnames and self._reader._has(field_name):
            field = self._reader.schema.fields[field_name]
            return _wrap(getattr(self._reader, name), self._streamdir, self._setdir, field=field)

        elif name == 'id' and self._reader._has('id0') and self._reader._has('id1'):
            return SetID(self._reader.id0, self._reader.id1)

        return getattr(self._reader, name)  # Not a field, but e.g. a method

    def __repr__(self):
        return '<Wrapper {}>'.format(self._reader.schema.node.displayName)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import time

import flask

from marv.model import Comment, Dataset, db
from .tooling import api_endpoint as marv_api_endpoint


@marv_api_endpoint('/comment', methods=['POST'])
def comment():
    username = flask.request.username
    changes = flask.request.get_json()
    if not changes:
        flask.abort(400)

    for id, ops in changes.items():
        now = int(time.time() * 1000)
        comments = [Comment(dataset_id=id, author=username, time_added=now, text=text)
                    for text in ops.get('add', [])]
        db.session.add_all(comments)
    db.session.commit()

    # TODO: inform about unknown setids?

    return flask.jsonify({})



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from marv_node.io import create_group
from marv_node.io import create_stream
from marv_node.io import fork
from marv_node.io import get_logger
from marv_node.io import get_requested
from marv_node.io import get_stream
from marv_node.io import make_file
from marv_node.io import pull
from marv_node.io import pull_all
from marv_node.io import pull_any
from marv_node.io import push
from marv_node.io import set_header
from marv_node.node import input
from marv_node.node import node
from marv_node.tools import select



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from collections import OrderedDict, namedtuple

import flask_sqlalchemy
from sqlalchemy import Index, event, schema
from sqlalchemy.engine import Engine
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.hybrid import hybrid_property

from marv_node.setid import SetID
from .utils import underscore_to_camelCase

_LISTING_PREFIX = ''

STATUS = OrderedDict((
    ('error', 'ERROR: One or more errors occured while processing this dataset'),
    ('missing', 'MISSING: One or more files of this dataset are missing'),
    ('outdated', 'OUTDATED: There are outdated nodes for this dataset'),
    ('pending', 'PENDING: There are pending node runs'),
))
STATUS_ERROR = 1
STATUS_MISSING = 2
STATUS_OUTDATED = 4
STATUS_PENDING = 8


db = flask_sqlalchemy.SQLAlchemy()
Boolean = db.Boolean
Column = db.Column
Float = db.Float
ForeignKey = db.ForeignKey
Integer = db.Integer
Model = db.Model
String = db.String
Table = db.Table
relationship = db.relationship


@event.listens_for(Engine, "connect")
def set_sqlite_pragma_connect(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    # page_size must be before journal_mode
    cursor.execute('PRAGMA foreign_keys=ON;')
    cursor.execute('PRAGMA page_size=4096;')
    cursor.execute('PRAGMA journal_mode=WAL')
    cursor.execute('PRAGMA synchronous=NORMAL;')
    # TODO: we don't want the websever to checkpoint,
    # ever. Checkpointing should happen via cli.
    cursor.close()


@event.listens_for(Engine, "close")
def set_sqlite_pragma_close(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute('PRAGMA optimize;')  # ignored prior to 3.18.0
    cursor.close()


@compiles(schema.CreateTable)
def compile(element, compiler, **kw):
    text = compiler.visit_create_table(element, **kw)
    if element.element.info.get('without_rowid'):
        text = text.rstrip() + ' WITHOUT ROWID\n\n'
    return text


def make_status_property(bitmask, doc=None):
    def fget(obj):
        return obj.status & bitmask

    def fset(obj, value):
        status = obj.status or 0
        if value:
            obj.status = status | bitmask
        else:
            obj.status = status - (status & bitmask)

    def fdel(obj):
        fset(obj, False)

    return property(fget, fset, fdel, doc)


dataset_tag = Table(
    'dataset_tag',
    Column('dataset_id', Integer, ForeignKey('dataset.id'), primary_key=True),
    Column('tag_id', Integer, ForeignKey('tag.id'), primary_key=True),
    info={'without_rowid': True},
)


class Dataset(Model):
    id = Column(Integer, primary_key=True)

    # TODO: consider Collection model
    collection = Column(String)
    discarded = Column(Boolean, default=False)
    name = Column(String)
    status = Column(Integer, default=0)
    time_added = Column(Integer, nullable=False)  # ms since epoch
    timestamp = Column(Integer, nullable=False)  # ms since epoch

    _setid = Column('setid', String, unique=True, nullable=False)
    @hybrid_property
    def setid(self):
        return SetID(self._setid)
    @setid.setter
    def setid(self, value):
        self._setid = str(value)
    @setid.expression
    def setid(self):
        return self._setid

    comments = relationship('Comment', back_populates='dataset', lazy='raise')
    files = relationship('File', back_populates='dataset', lazy='joined')
    # Handled via backref on Tag
    # tags = relationship('Tag', secondary=dataset_tag, lazy='raise',
    #                        back_populates='datasets'),

    error = make_status_property(1)
    missing = make_status_property(2)
    outdated = make_status_property(4)
    pending = make_status_property(8)

    def __repr__(self):
        return "<{} {} {}>".format(type(self).__name__, self.setid, self.name.encode('utf-8'))


class File(Model):
    __table_args__ = (
        {'info': {'without_rowid': True}},
    )
    dataset_id = Column(Integer, ForeignKey('dataset.id'), primary_key=True)
    idx = Column(Integer, primary_key=True)  # files are counted per dataset

    missing = Column(Boolean, default=False)
    mtime = Column(Integer, nullable=False)  # ms since epoch
    path = Column(String, nullable=False)
    size = Column(Integer, nullable=False)

    dataset = relationship('Dataset', back_populates='files', lazy='raise')

    def __repr__(self):
        return "<{} '{}'>".format(type(self).__name__, self.path.encode('utf-8'))


class Comment(Model):
    id = Column(Integer, primary_key=True)
    dataset_id = Column(Integer, ForeignKey('dataset.id'))

    author = Column(String, nullable=False)
    time_added = Column(Integer, nullable=False)  # ms since epoch
    text = Column(String, nullable=False)

    dataset = relationship('Dataset', back_populates='comments', lazy='raise')


class Tag(Model):
    __table_args__ = (
        Index('idx_tag_collection_value', 'collection', 'value', unique=True),
    )
    id = Column(Integer, primary_key=True)
    collection = Column(String, nullable=False)
    value = Column(String)
    datasets = relationship('Dataset', secondary=dataset_tag, lazy='raise',
                            backref='tags')


user_group = Table(
    'user_group',
    Column('user_id', Integer, ForeignKey('user.id'), primary_key=True),
    Column('group_id', Integer, ForeignKey('group.id'), primary_key=True),
    info={'without_rowid': True},
)


class User(Model):
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False, unique=True)
    password = Column(String)
    given_name = Column(String)
    family_name = Column(String)
    email = Column(String)
    realm = Column(String, nullable=False)
    realmuid = Column(String)
    time_created = Column(Integer, nullable=False)
    time_updated = Column(Integer, nullable=False)


class Group(Model):
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False, unique=True)
    # TODO: switch to lazy raise
    users = relationship('User', secondary=user_group, backref='groups')


dataset = Dataset.__table__
file = File.__table__
comment = Comment.__table__
tag = Tag.__table__
user = User.__table__
group = Group.__table__


ListingModel = namedtuple('ListingModel', 'db Listing relations secondaries')

def make_listing_model(name, filter_specs):
    relations = {}
    secondaries = {}
    listing_name = '{}listing_{}'.format(_LISTING_PREFIX, name)
    listing_model_name = underscore_to_camelCase(listing_name).encode('ascii')
    listingid = '{}.id'.format(listing_name)

    def generate_relation(fname):
        sec_name = '{}_{}_sec'.format(listing_name, fname)
        rel_name = '{}_{}'.format(listing_name, fname)
        rel_model_name = underscore_to_camelCase(rel_name).encode('ascii')
        relid = '{}.id'.format(rel_name)
        assert not '_' in rel_model_name, rel_model_name
        secondary = Table(
            sec_name,
            Column('listing_id', Integer, ForeignKey(listingid), primary_key=True),
            Column('relation_id', Integer, ForeignKey(relid), primary_key=True),
            info={'without_rowid': True},
        )
        secondaries[fname] = secondary
        relations[fname] = type(rel_model_name, (Model,), {
            'id': Column(Integer, primary_key=True),
            'value': Column(String, index=True, unique=True),
            'listing': relationship(listing_model_name,
                                    secondary=secondary,
                                    lazy='raise',
                                    back_populates=fname),
            '__repr__': lambda self: "<{} {} {}'>".format(
                type(self).__name__, self.id, self.value.encode('utf-8'))
        })
        return relationship(rel_model_name, secondary=secondary,
                            lazy='raise', back_populates='listing')

    coltype_factories = {
        # All dates and times in ms (since epoch)
        'datetime': lambda name: Column(Integer),
        'filesize': lambda name: Column(Integer),
        'float': lambda name: Column(Float),
        'int': lambda name: Column(Integer),
        'string': lambda name: Column(String),
        'string[]': generate_relation,
        'subset': generate_relation,
        'timedelta': lambda name: Column(Integer),
        'words': lambda name: Column(String),
    }

    dct = {
        '__table_args__': (
            {'info': {'without_rowid': True}}
        ),
        'id': Column(Integer, ForeignKey('dataset.id'), primary_key=True),
        'row': Column(String),
        'dataset': relationship('Dataset'),
        '__repr__': lambda self: "<{} {}>".format(type(self).__name__, self.id)
    }
    assert not filter_specs.viewkeys() & dct.viewkeys()
    dct.update((fspec.name, coltype_factories[fspec.value_type](fspec.name))
               for fspec in filter_specs.values()
               if fspec.name not in ('comments', 'status', 'tags'))

    Listing = type(listing_model_name, (Model,), dct)
    return ListingModel(db, Listing, relations, secondaries)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import json
import os
from itertools import count

import mock
import pytest
import sqlalchemy as sqla
from flask import current_app
from pathlib2 import Path

import marv
import marv.app
from marv.scanner import DatasetInfo
from marv.site import Site, dump_database
from marv.types import Int8Value, Section
from marv_nodes import dataset as dataset_node


DATADIR = Path(__file__).parent / 'data'
RECORD = os.environ.get('MARV_TESTING_RECORD')
SETIDS = [
    'l2vnfhfoe3z7ad7vclkd64tsqy',
    'vdls3sgw5yanuat4uepmhjwcpm',
    'e2oxlhpedjwked2llnnzir4tii',
    'phjg4ncymwbrbl4yx35bciqazq'
]

MARV_CONF = """
[marv]
acl = marv_webapi.acls:public
collections = foo bar

[collection foo]
scanner = marv.tests.test_dump_restore:scanner
scanroots = ./scanroots/foo
nodes =
    marv_nodes:dataset
    marv_nodes:meta_table
    marv_nodes:summary_keyval
    marv.tests.test_dump_restore:node_test
    marv.tests.test_dump_restore:section_test
filters =
    name       | Name       | substring         | string   | (get "dataset.name")
    setid      | Set Id     | startswith        | string   | (get "dataset.id")
    size       | Size       | lt le eq ne ge gt | filesize | (sum (get "dataset.files[:].size"))
    status     | Status     | any all           | subset   | (status )
    tags       | Tags       | any all           | subset   | (tags )
    comments   | Comments   | substring         | string   | (comments )
    files      | File paths | substring_any     | string[] | (get "dataset.files[:].path")
    time_added | Added      | lt le eq ne ge gt | datetime | (get "dataset.time_added")
    node_test  | Test node  | any               | subset   | (list (get "node_test.value" 0))
listing_columns =
    name       | Name   | route    | (detail_route (get "dataset.id") (get "dataset.name"))
    size       | Size   | filesize | (sum (get "dataset.files[:].size"))
    status     | Status | icon[]   | (status )
    tags       | Tags   | pill[]   | (tags )
    time_added | Added  | datetime | (get "dataset.time_added")
    node_test  | Test   | int      | (get "node_test.value" 0)
detail_sections =
    section_test

[collection bar]
scanner = marv.tests.test_dump_restore:scanner
scanroots = ./scanroots/bar
nodes =
    marv_nodes:dataset
    marv_nodes:meta_table
    marv_nodes:summary_keyval
    marv.tests.test_dump_restore:node_test
    marv.tests.test_dump_restore:section_test
filters =
    name       | Name       | substring         | string   | (get "dataset.name")
    setid      | Set Id     | startswith        | string   | (get "dataset.id")
    size       | Size       | lt le eq ne ge gt | filesize | (sum (get "dataset.files[:].size"))
    status     | Status     | any all           | subset   | (status )
    tags       | Tags       | any all           | subset   | (tags )
    comments   | Comments   | substring         | string   | (comments )
    files      | File paths | substring_any     | string[] | (get "dataset.files[:].path")
    time_added | Added      | lt le eq ne ge gt | datetime | (get "dataset.time_added")
    node_test  | Test node  | any               | subset   | (list (get "node_test.value" 0))
listing_columns =
    name       | Name   | route    | (detail_route (get "dataset.id") (get "dataset.name"))
    size       | Size   | filesize | (sum (get "dataset.files[:].size"))
    status     | Status | icon[]   | (status )
    tags       | Tags   | pill[]   | (tags )
    time_added | Added  | datetime | (get "dataset.time_added")
    node_test  | Test   | int      | (get "node_test.value" 0)
detail_sections =
    section_test
"""


def scanner(dirpath, dirnames, filenames):
    return [DatasetInfo(x, [x]) for x in filenames]


@marv.node(Int8Value)
@marv.input('dataset', default=dataset_node)
def node_test(dataset):
    dataset = yield marv.pull(dataset)
    with open(dataset.files[0].path) as f:
        yield marv.push({'value': int(f.read())})


@marv.node(Section)
@marv.input('node_test', default=node_test)
def section_test(node_test):
    value = yield marv.pull(node_test)
    value = value.value
    yield marv.push({'title': 'Test', 'widgets': [
        {'keyval': {'items': [{'title': 'value', 'cell': {'uint64': value}}]}}
    ]})


@pytest.fixture(scope='function')
def site(tmpdir):
    flag = (tmpdir / 'TEST_SITE')
    flag.write('')

    marv_conf = (tmpdir / 'marv.conf')
    marv_conf.write(MARV_CONF)

    # make scanroots
    for sitename in ('foo', 'bar'):
        for idx, name in enumerate(['a', u'\u03a8'.encode('utf-8')]):
            name = '{}_{}'.format(sitename, name)
            path = tmpdir / 'scanroots' / sitename / name
            path.write(str(idx), ensure=True)

    yield Site(marv_conf.strpath)


@pytest.fixture(scope='function')
def app(site):
    app = marv.app.create_app(site, init=True)
    app.testing = True
    with app.app_context():
        app = app.test_client()
        def get_json(*args, **kw):
            resp = app.get(*args, **kw)
            return json.loads(resp.data)
        app.get_json = get_json
        yield app


def recorded(data, filename):
    """Assert data corresponds to recorded data.

    Record data if RECORD is set. Return True on success to enable
    being used as ``assert recorded()`` for nicer readability.

    """
    json_file = DATADIR / filename
    if RECORD:
        json_file.write_bytes(json.dumps(data, sort_keys=True, indent=2))
    recorded_data = json.loads(json_file.read_bytes())
    assert recorded_data == data, filename
    return True


def test_dump(app, site):
    sitedir = os.path.dirname(site.config.filename)

    # Ensure all tables are empty / at factory defaults
    select = sqla.sql.select
    engine = sqla.create_engine(site.config.marv.dburi)
    meta = sqla.MetaData(engine)
    meta.reflect()
    con = engine.connect()
    tables = {k: v for k, v in meta.tables.items() if not k.startswith('listing_')}
    assert tables.viewkeys() == \
        {'dataset', 'dataset_tag', 'tag', 'file', 'comment', 'user', 'user_group', 'group'}
    for name, table in sorted(tables.items()):
        rows = list(con.execute(select([table])))
        if name == 'group':
            assert rows == [(1, 'admin')]
        else:
            assert not rows, name

    dump = dump_database(site.config.marv.dburi)
    assert recorded(dump, 'empty_dump.json')

    metadata = app.get_json('/marv/api/meta')
    listings = {}
    for colinfo in metadata['collections']['items']:
        name = colinfo['id']
        listings[name] = app.get_json('/marv/api/collection/{}'.format(name))
    assert recorded(listings, 'empty_listings.json')

    # Populate database, asserting in multiple ways that it is populated
    with mock.patch('marv_node.setid.SetID.random', side_effect=SETIDS) as _, \
         mock.patch('marv.utils.time', side_effect=count(2000)) as __, \
         mock.patch('marv.utils.mtime', side_effect=count(1000)):
        site.scan()
    um = current_app.um

    with mock.patch('bcrypt.gensalt', return_value='$2b$12$k67acf6S32i3nW0c7ycwe.') as _, \
         mock.patch('marv.utils.time', side_effect=count(2000)):
        um.user_add('user1', 'pw1', 'marv', '')
        um.user_add('user2', 'pw2', 'marv', '')
    um.group_add('grp')
    um.group_adduser('admin', 'user1')
    um.group_adduser('grp', 'user2')

    fooids = site.query(['foo'])
    barids = site.query(['bar'])
    site.tag(fooids, add=['TAG1'])
    site.tag([fooids[1], barids[1]], add=['TAG2'])

    fooid0 = con.execute('SELECT id FROM dataset WHERE setid = $1', fooids[0]).scalar()
    fooid1 = con.execute('SELECT id FROM dataset WHERE setid = $1', fooids[1]).scalar()
    barid1 = con.execute('SELECT id FROM dataset WHERE setid = $1', barids[1]).scalar()
    with mock.patch('marv.utils.now', side_effect=count(100000)):
        site.comment('user1', 'comment\ntext', [fooid0])
        site.comment('user2', 'more\ncomment', [fooid1, barid1])

    # Ensure all tables have been populated
    tables = {k: v for k, v in meta.tables.items() if not k.startswith('listing_')}
    for name, table in sorted(tables.items()):
        rows = list(con.execute(select([table])))
        if name == 'group':
            assert len(rows) > 1
        else:
            assert rows, name

    # Run nodes
    for setid in fooids + barids:
        changed = site.run(setid)
        assert len(changed) == 5
        changed = site.run(setid)
        assert not changed

    listings = {}
    for colinfo in metadata['collections']['items']:
        name = colinfo['id']
        listings[name] = lst = []
        data = app.get_json('/marv/api/collection/{}'.format(name))
        rows = data['listing']['widget']['data']['rows']
        for row in sorted(rows, key=lambda x: x['setid']):
            del row['id']
            lst.append(row)
    assert recorded(listings, 'full_listings.json')

    details = []
    for colname, rows in sorted(listings.items()):
        for row in rows:
            detail = app.get_json('/marv/api/dataset/{}'.format(row['setid']))
            del detail['id']
            assert detail['collection'] == colname
            details.append(detail)
    assert recorded(details, 'full_details.json')

    # Dump database
    dump = dump_database(site.config.marv.dburi)
    for datasets in dump['datasets'].values():
        for ds in datasets:
            for file in ds['files']:
                file['path'] = file['path'].replace(sitedir, 'SITEDIR')
    assert recorded(dump, 'full_dump.json')


def test_restore(app, site):
    sitedir = os.path.dirname(site.config.filename)

    # Ensure all tables are empty / at factory defaults
    select = sqla.sql.select
    engine = sqla.create_engine(site.config.marv.dburi)
    meta = sqla.MetaData(engine)
    meta.reflect()
    con = engine.connect()
    tables = {k: v for k, v in meta.tables.items() if not k.startswith('listing_')}
    assert tables.viewkeys() == \
        {'dataset', 'dataset_tag', 'tag', 'file', 'comment', 'user', 'user_group', 'group'}
    for name, table in sorted(tables.items()):
        rows = list(con.execute(select([table])))
        if name == 'group':
            assert rows == [(1, 'admin')]
        else:
            assert not rows, name

    dump = dump_database(site.config.marv.dburi)
    assert recorded(dump, 'empty_dump.json')

    metadata = app.get_json('/marv/api/meta')
    listings = {}
    for colinfo in metadata['collections']['items']:
        name = colinfo['id']
        listings[name] = app.get_json('/marv/api/collection/{}'.format(name))
    assert recorded(listings, 'empty_listings.json')

    # Restore database
    full_dump = json.loads((DATADIR / 'full_dump.json').read_bytes())
    for datasets in full_dump['datasets'].values():
        for ds in datasets:
            for file in ds['files']:
                file['path'] = file['path'].replace('SITEDIR', sitedir)
    site.restore_database(**full_dump)

    # Ensure all tables have been populated
    tables = {k: v for k, v in meta.tables.items() if not k.startswith('listing_')}
    for name, table in sorted(tables.items()):
        rows = list(con.execute(select([table])))
        if name == 'group':
            assert len(rows) > 1
        else:
            assert rows, name

    # Run nodes
    fooids = site.query(['foo'])
    barids = site.query(['bar'])
    for setid in fooids + barids:
        changed = site.run(setid)
        assert len(changed) == 5
        changed = site.run(setid)
        assert not changed

    # Assert listing is still the same
    listings = {}
    for colinfo in metadata['collections']['items']:
        name = colinfo['id']
        listings[name] = lst = []
        data = app.get_json('/marv/api/collection/{}'.format(name))
        rows = data['listing']['widget']['data']['rows']
        for row in sorted(rows, key=lambda x: x['setid']):
            del row['id']
            lst.append(row)
    assert recorded(listings, 'full_listings.json')

    # Assert detail is still the same
    details = []
    for colname, rows in sorted(listings.items()):
        for row in rows:
            detail = app.get_json('/marv/api/dataset/{}'.format(row['setid']))
            del detail['id']
            assert detail['collection'] == colname
            details.append(detail)
    assert recorded(details, 'full_details.json')

    # Redump database and assert dumps are the same
    dump = dump_database(site.config.marv.dburi)
    dump = json.loads(json.dumps(dump))
    full_dump = json.loads((DATADIR / 'full_dump.json').read_bytes())
    for datasets in full_dump['datasets'].values():
        for ds in datasets:
            for file in ds['files']:
                file['path'] = file['path'].replace('SITEDIR', sitedir)
    assert full_dump == dump



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function


from collections import OrderedDict


class DefaultOrderedDict(OrderedDict):
    def __init__(self, factory):
        super(DefaultOrderedDict, self).__init__()
        self._factory = factory

    def __getitem__(self, key):
        try:
            return super(DefaultOrderedDict, self).__getitem__(key)
        except KeyError:
            value = self._factory()
            self[key] = value
            return value
